from bbot.modules.base import BaseModule


class aggregate(BaseModule):
    watched_events = ["SUMMARY"]
    produced_events = ["SUMMARY"]
    flags = ["passive", "safe"]
    meta = {"description": "Report on scan statistics"}

    def report(self):
        for table_row in str(self.scan.stats).splitlines():
            self.info(table_row)



from bbot.modules.shodan_dns import shodan_dns


class fullhunt(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the fullhunt.io API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "FullHunt API Key"}

    base_url = "https://fullhunt.io/api/v1"

    def setup(self):
        self.api_key = self.config.get("api_key", "")
        self.headers = {"x-api-key": self.api_key}
        return super().setup()

    def ping(self):
        url = f"{self.base_url}/auth/status"
        j = self.helpers.request(url, headers=self.headers).json()
        remaining = j["user_credits"]["remaining_credits"]
        assert remaining > 0, "No credits remaining"

    def request_url(self, query):
        url = f"{self.base_url}/domain/{self.helpers.quote(query)}/subdomains"
        return self.helpers.request(url, headers=self.headers)

    def parse_results(self, r, query):
        return r.json().get("hosts", [])



import logging

log = logging.getLogger("bbot.scanner.stats")


def _increment(d, k):
    try:
        d[k] += 1
    except KeyError:
        d[k] = 1


class ScanStats:
    def __init__(self, scan):
        self.scan = scan
        self.module_stats = {}
        self.events_emitted_by_type = {}
        self.perf_stats = []

    def function_called(self, qualname, runtime):
        # uncomment the line below to track durations for function calls
        # this is helpful for debugging elusive performance issues
        # self.perf_stats.append((qualname, runtime))
        pass

    def event_distributed(self, event):
        _increment(self.events_emitted_by_type, event.type)
        module_stat = self.get(event.module)
        if module_stat is not None:
            module_stat.increment_emitted(event)

    def event_produced(self, event):
        module_stat = self.get(event.module)
        if module_stat is not None:
            module_stat.increment_produced(event)

    def event_consumed(self, event, module):
        module_stat = self.get(module)
        if module_stat is not None:
            module_stat.increment_consumed(event)

    def get(self, module):
        try:
            module_stat = self.module_stats[module.name]
        except KeyError:
            module_stat = ModuleStat(module)
            self.module_stats[module.name] = module_stat
        except AttributeError:
            module_stat = None
        return module_stat

    def table(self):
        header = ["Module", "Produced", "Consumed"]
        table = []
        for mname, mstat in self.module_stats.items():
            if mname == "TARGET" or mstat.module._stats_exclude:
                continue
            table_row = []
            table_row.append(mname)
            produced_str = f"{mstat.produced_total:,}"
            produced = sorted(mstat.produced.items(), key=lambda x: x[0])
            if produced:
                produced_str += " (" + ", ".join(f"{c:,} {t}" for t, c in produced) + ")"
            table_row.append(produced_str)
            consumed_str = f"{mstat.consumed_total:,}"
            consumed = sorted(mstat.consumed.items(), key=lambda x: x[0])
            if consumed:
                consumed_str += " (" + ", ".join(f"{c:,} {t}" for t, c in consumed) + ")"
            table_row.append(consumed_str)
            table.append(table_row)
        table.sort(key=lambda x: self.module_stats[x[0]].produced_total, reverse=True)
        return [header] + table

    def __str__(self):
        self.perf_stats.sort(key=lambda x: x[-1])
        for callback, runtime in self.perf_stats:
            log.info(f"{callback}\t{runtime}")
        table = self.table()
        if len(table) == 1:
            table += [["None", "None", "None"]]
        return self.scan.helpers.make_table(table[1:], table[0])


class ModuleStat:
    def __init__(self, module):
        self.module = module

        self.emitted = {}
        self.emitted_total = 0
        self.produced = {}
        self.produced_total = 0
        self.consumed = {}
        self.consumed_total = 0

    def increment_emitted(self, event):
        self.emitted_total += 1
        _increment(self.emitted, event.type)

    def increment_produced(self, event):
        self.produced_total += 1
        _increment(self.produced, event.type)

    def increment_consumed(self, event):
        if event.type not in ("FINISHED",):
            self.consumed_total += 1
            _increment(self.consumed, event.type)



from .header_brute import header_brute
from bbot.core.errors import ScanCancelledError


class getparam_brute(header_brute):
    """
    Inspired by https://github.com/PortSwigger/param-miner
    """

    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["brute-force", "active", "aggressive", "slow", "web-paramminer"]
    meta = {"description": "Check for common HTTP GET parameters"}

    options = {"wordlist": "https://raw.githubusercontent.com/PortSwigger/param-miner/master/resources/params"}
    options_desc = {"wordlist": "Define the wordlist to be used to derive GET params"}
    scanned_hosts = []
    getparam_blacklist = []
    max_threads = 12
    in_scope_only = True
    compare_mode = "getparam"

    def check_batch(self, compare_helper, url, getparam_list):
        if self.scan.stopping:
            raise ScanCancelledError()
        test_getparams = {p: self.rand_string(14) for p in getparam_list}
        return compare_helper.compare(self.helpers.add_get_params(url, test_getparams).geturl())

    def gen_count_args(self, url):
        getparam_count = 40
        while 1:
            if getparam_count < 0:
                break
            fake_getparams = {self.rand_string(14): self.rand_string(14) for _ in range(0, getparam_count)}
            yield getparam_count, (self.helpers.add_get_params(url, fake_getparams).geturl(),), {}
            getparam_count -= 5

    def clean_list(self, getparam):
        if (len(getparam) > 0) and (getparam.strip() not in self.getparam_blacklist):
            return True
        return False



from ..bbot_fixtures import *  # noqa: F401


def test_cloud_helpers(monkeypatch, bbot_scanner, bbot_config):
    scan1 = bbot_scanner("127.0.0.1", config=bbot_config)
    scan1.load_modules()
    aws_event1 = scan1.make_event("amazonaws.com", source=scan1.root_event)
    aws_event2 = scan1.make_event("asdf.amazonaws.com", source=scan1.root_event)
    aws_event3 = scan1.make_event("asdfamazonaws.com", source=scan1.root_event)
    providers = scan1.helpers.cloud.providers

    # make sure they're all here
    assert "aws" in providers
    assert "gcp" in providers
    assert "azure" in providers
    assert "digitalocean" in providers

    # make sure tagging is working
    aws = providers["aws"]
    aws.tag_event(aws_event1)
    assert "cloud-aws" in aws_event1.tags
    aws.tag_event(aws_event2)
    assert "cloud-aws" in aws_event2.tags
    aws.tag_event(aws_event3)
    assert not "cloud-aws" in aws_event3.tags

    # test storage bucket extraction
    storage_bucket_hosts = [
        "asdf.s3-asdf.amazonaws.com",
        "asdf.storage.googleapis.com",
        "asdf.blob.core.windows.net",
        "asdf.digitaloceanspaces.com",
    ]
    dummy_body = ""
    for h in storage_bucket_hosts:
        dummy_body += f'<a src="https://{h}"/>'
    dummy_response = {"body": dummy_body, "url": "http://example.com"}
    http_response = scan1.make_event(dummy_response, "HTTP_RESPONSE", source=scan1.root_event)
    results = []
    for provider_name, provider in providers.items():
        monkeypatch.setattr(
            provider,
            "emit_event",
            lambda *args, **kwargs: results.append(kwargs.get("data", {}).get("url", "")),
        )
        provider.excavate(http_response)
    for h in storage_bucket_hosts:
        assert f"https://{h}" in results, f"cloud helpers failed to excavate {h} from {dummy_body}"



from contextlib import suppress

from censys.common import exceptions
from censys.search import CensysHosts
from censys.search import CensysCertificates

from bbot.modules.shodan_dns import shodan_dns


class censys(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME", "EMAIL_ADDRESS", "IP_ADDRESS", "OPEN_PORT", "PROTOCOL"]
    flags = ["subdomain-enum", "email-enum", "passive", "safe"]
    meta = {"description": "Query the Censys API", "auth_required": True}
    options = {"api_id": "", "api_secret": "", "max_records": 1000}
    options_desc = {
        "api_id": "Censys.io API ID",
        "api_secret": "Censys.io API Secret",
        "max_records": "Limit results to help prevent exceeding API quota",
    }

    deps_pip = ["censys"]

    def setup(self):
        self.max_records = self.config.get("max_records", 1000)
        self.api_id = self.config.get("api_id", "")
        self.api_secret = self.config.get("api_secret", "")
        self._cert_name_threshold = 20
        with suppress(Exception):
            self.hosts = CensysHosts(api_id=self.api_id, api_secret=self.api_secret)
        with suppress(Exception):
            self.certificates = CensysCertificates(api_id=self.api_id, api_secret=self.api_secret)
        return super().setup()

    def ping(self):
        quota = self.certificates.quota()
        used = int(quota["used"])
        allowance = int(quota["allowance"])
        assert used < allowance, "No quota remaining"

    def query(self, query):
        emails = set()
        dns_names = set()
        ip_addresses = dict()
        try:
            # certificates
            certificate_query = f"parsed.names: {query}"
            certificate_fields = ["parsed.names", "parsed.issuer_dn", "parsed.subject_dn"]
            for result in self.certificates.search(
                certificate_query, fields=certificate_fields, max_records=self.max_records
            ):
                parsed_names = result.get("parsed.names", [])
                # helps filter out third-party certs with a lot of garbage names
                _filter = lambda x: True
                domain = self.helpers.tldextract(query).domain
                if len(parsed_names) > self._cert_name_threshold:
                    _filter = lambda x: domain in str(x.lower())
                parsed_names = list(filter(_filter, parsed_names))
                dns_names.update(set([n.lstrip(".*").rstrip(".").lower() for n in parsed_names]))
                emails.update(set(self.helpers.extract_emails(result.get("parsed.issuer_dn", ""))))
                emails.update(set(self.helpers.extract_emails(result.get("parsed.subject_dn", ""))))

            # hosts
            per_page = 100
            pages = max(1, int(self.max_records / per_page))
            hosts_query = f"services.tls.certificates.leaf_data.names: {query} or services.tls.certificates.leaf_data.subject.email_address: {query}"
            for i, page in enumerate(self.hosts.search(hosts_query, per_page=per_page, pages=pages)):
                for result in page:
                    ip = result.get("ip", "")
                    if not ip:
                        continue
                    ip_addresses[ip] = []
                    services = result.get("services", [])
                    for service in services:
                        port = service.get("port")
                        service_name = service.get("service_name", "")
                        transport_protocol = service.get("transport_protocol", "")
                        if not port or not transport_protocol:
                            continue
                        ip_addresses[ip].append((port, service_name, transport_protocol))
                if self.scan.stopping:
                    break

        except exceptions.CensysRateLimitExceededException:
            self.warning("Exceeded Censys account limits")
        except exceptions.CensysException as e:
            self.warning(f"Error with API: {e}")
        except Exception as e:
            self.warning(f"Unknown error: {e}")

        return emails, dns_names, ip_addresses

    def handle_event(self, event):
        query = self.make_query(event)
        emails, dns_names, ip_addresses = self.query(query)
        for email in emails:
            self.emit_event(email, "EMAIL_ADDRESS", source=event)
        for dns_name in dns_names:
            self.emit_event(dns_name, "DNS_NAME", source=event)
        for ip, services in ip_addresses.items():
            ip_event = self.make_event(ip, "IP_ADDRESS", source=event)
            if not ip_event:
                continue
            self.emit_event(ip_event)
            for port, service_name, transport_protocol in services:
                port_data = self.helpers.make_netloc(ip, port)
                port_type = f"OPEN_{transport_protocol.upper()}_PORT"
                port_event = self.make_event(port_data, port_type, source=ip_event)
                if not port_event:
                    continue
                self.emit_event(port_event)
                if service_name:
                    service_name = str(service_name).upper()
                    protocol_data = {"host": port_data, "protocol": service_name}
                    self.emit_event(protocol_data, "PROTOCOL", source=port_event)

    @property
    def auth_secret(self):
        return self.api_id and self.api_secret



import os
import sys
from pathlib import Path
from omegaconf import OmegaConf

from . import files, args, environ
from ..errors import ConfigLoadError
from ...modules import module_loader
from ..helpers.misc import mkdir, error_and_exit, filter_dict, clean_dict, log_to_stderr

# cached sudo password
bbot_sudo_pass = None

try:
    config = OmegaConf.merge(
        # first, pull module defaults
        OmegaConf.create(
            {
                "modules": module_loader.configs(type="scan"),
                "output_modules": module_loader.configs(type="output"),
                "internal_modules": module_loader.configs(type="internal"),
            }
        ),
        # then look in .yaml files
        files.get_config(),
        # finally, pull from CLI arguments
        args.get_config(),
    )
except ConfigLoadError as e:
    error_and_exit(e)


config = environ.prepare_environment(config)


# ensure bbot.yml
if not files.config_filename.exists():
    log_to_stderr(f"Creating BBOT config at {files.config_filename}")
    no_secrets_config = OmegaConf.to_object(config)
    no_secrets_config = clean_dict(no_secrets_config, "api_key", "username", "password", "token", fuzzy=True)
    yaml = OmegaConf.to_yaml(no_secrets_config)
    yaml = "\n".join(f"# {line}" for line in yaml.splitlines())
    with open(str(files.config_filename), "w") as f:
        f.write(yaml)

# ensure secrets.yml
if not files.secrets_filename.exists():
    log_to_stderr(f"Creating BBOT secrets at {files.secrets_filename}")
    secrets_only_config = OmegaConf.to_object(config)
    secrets_only_config = filter_dict(
        secrets_only_config, "api_key", "username", "password", "token", "secret", "_id", fuzzy=True
    )
    yaml = OmegaConf.to_yaml(secrets_only_config)
    yaml = "\n".join(f"# {line}" for line in yaml.splitlines())
    with open(str(files.secrets_filename), "w") as f:
        f.write(yaml)
    files.secrets_filename.chmod(0o600)



import logging
import requests
from time import sleep
from pathlib import Path
from requests_cache import CachedSession
from requests_cache.backends import SQLiteCache
from requests.exceptions import RequestException

from bbot.core.errors import WordlistError, CurlError

log = logging.getLogger("bbot.core.helpers.web")


def wordlist(self, path, lines=None, **kwargs):
    if not path:
        raise WordlistError(f"Invalid wordlist: {path}")
    if not "cache_hrs" in kwargs:
        kwargs["cache_hrs"] = 720
    if self.is_url(path):
        filename = self.download(str(path), **kwargs)
        if filename is None:
            raise WordlistError(f"Unable to retrieve wordlist from {path}")
    else:
        filename = Path(path).resolve()
        if not filename.is_file():
            raise WordlistError(f"Unable to find wordlist at {path}")

    if lines is None:
        return filename
    else:
        lines = int(lines)
        with open(filename) as f:
            read_lines = f.readlines()
        cache_key = f"{filename}:{lines}"
        truncated_filename = self.cache_filename(cache_key)
        with open(truncated_filename, "w") as f:
            for line in read_lines[:lines]:
                f.write(line)
        return truncated_filename


def download(self, url, **kwargs):
    """
    Downloads file, returns full path of filename
    If download failed, returns None

    Caching supported via "cache_hrs"
    """
    success = False
    filename = self.cache_filename(url)
    cache_hrs = float(kwargs.pop("cache_hrs", -1))
    log.debug(f"Downloading file from {url} with cache_hrs={cache_hrs}")
    if cache_hrs > 0 and self.is_cached(url):
        log.debug(f"{url} is cached")
        success = True
    else:
        method = kwargs.get("method", "GET")
        try:
            with self.request(method=method, url=url, stream=True, raise_error=True, **kwargs) as response:
                status_code = getattr(response, "status_code", 0)
                log.debug(f"Download result: HTTP {status_code}")
                if status_code != 0:
                    response.raise_for_status()
                    with open(filename, "wb") as f:
                        for chunk in response.iter_content(chunk_size=8192):
                            f.write(chunk)
                    success = True
        except RequestException as e:
            log.warning(f"Failed to download {url}: {e}")
            return
        except AttributeError:
            return

    if success:
        return filename.resolve()


def request(self, *args, **kwargs):
    """
    Multipurpose function for making web requests

    Supports custom sessions
        session Request.Session()

    Arguments
        cache_for (Union[None, int, float, str, datetime, timedelta]): Cache response for <int> seconds
        raise_error (bool): Whether to raise exceptions (default: False)
    """

    raise_error = kwargs.pop("raise_error", False)

    cache_for = kwargs.pop("cache_for", None)
    if cache_for is not None:
        log.debug(f"Caching HTTP session with expire_after={cache_for}")
        try:
            session = self.cache_sessions[cache_for]
        except KeyError:
            db_path = str(self.cache_dir / "requests-cache.sqlite")
            backend = SQLiteCache(db_path=db_path)
            session = CachedSession(expire_after=cache_for, backend=backend)
            self.cache_sessions[cache_for] = session

    if kwargs.pop("session", None) or not cache_for:
        session = kwargs.pop("session", None)

    http_timeout = self.config.get("http_timeout", 20)
    user_agent = self.config.get("user_agent", "BBOT")

    # in case of URL only, assume GET request
    if len(args) == 1:
        kwargs["url"] = args[0]
        args = []

    url = kwargs.get("url", "")
    retries = kwargs.pop("retries", 0)

    if not args and "method" not in kwargs:
        kwargs["method"] = "GET"

    if not "timeout" in kwargs:
        kwargs["timeout"] = http_timeout

    headers = kwargs.get("headers", None)

    if headers is None:
        headers = {}
    if "User-Agent" not in headers:
        headers.update({"User-Agent": user_agent})
    # only add custom headers if the URL is in-scope
    if self.scan.in_scope(url):
        for hk, hv in self.scan.config.get("http_headers", {}).items():
            # don't clobber headers
            if hk not in headers:
                headers[hk] = hv
    kwargs["headers"] = headers

    http_debug = self.config.get("http_debug", False)
    while retries == "infinite" or retries >= 0:
        try:
            if http_debug:
                logstr = f"Web request: {str(args)}, {str(kwargs)}"
                log.debug(logstr)
            if session is not None:
                response = session.request(*args, **kwargs)
            else:
                response = requests.request(*args, **kwargs)
            if http_debug:
                log.debug(f"Web response: {response} (Length: {len(response.content)}) headers: {response.headers}")
            return response
        except RequestException as e:
            log.debug(f"Error with request: {e}")
            if retries != "infinite":
                retries -= 1
            if retries == "infinite" or retries >= 0:
                log.warning(f'Error requesting "{url}" ({e}), retrying...')
                sleep(1)
            else:
                if raise_error:
                    raise e


def api_page_iter(self, url, page_size=100, json=True, **requests_kwargs):
    page = 1
    offset = 0
    while 1:
        new_url = url.format(page=page, page_size=page_size, offset=offset)
        result = self.request(new_url, **requests_kwargs)
        try:
            if json:
                result = result.json()
            yield result
        except Exception:
            import traceback

            log.warning(f'Error in api_page_iter() for url: "{new_url}"')
            log.trace(traceback.format_exc())
            break
        finally:
            offset += page_size
            page += 1


def curl(self, *args, **kwargs):
    url = kwargs.get("url", "")

    if not url:
        raise CurlError("No URL supplied to CURL helper")

    curl_command = ["curl", url, "-s"]

    raw_path = kwargs.get("raw_path", False)
    if raw_path:
        curl_command.append("--path-as-is")

    # respect global ssl verify settings
    ssl_verify = self.config.get("ssl_verify")
    if ssl_verify == False:
        curl_command.append("-k")

    headers = kwargs.get("headers", {})

    ignore_bbot_global_settings = kwargs.get("ignore_bbot_global_settings", False)

    if ignore_bbot_global_settings:
        log.debug("ignore_bbot_global_settings enabled. Global settings will not be applied")
    else:
        http_timeout = self.config.get("http_timeout", 20)
        user_agent = self.config.get("user_agent", "BBOT")

        if "User-Agent" not in headers:
            headers["User-Agent"] = user_agent

        # only add custom headers if the URL is in-scope
        if self.scan.in_scope(url):
            for hk, hv in self.scan.config.get("http_headers", {}).items():
                headers[hk] = hv

        # add the timeout
        if not "timeout" in kwargs:
            timeout = http_timeout

        curl_command.append("-m")
        curl_command.append(str(timeout))

    for k, v in headers.items():
        if isinstance(v, list):
            for x in v:
                curl_command.append("-H")
                curl_command.append(f"{k}: {x}")

        else:
            curl_command.append("-H")
            curl_command.append(f"{k}: {v}")

    post_data = kwargs.get("post_data", {})
    if len(post_data.items()) > 0:
        curl_command.append("-d")
        post_data_str = ""
        for k, v in post_data.items():
            post_data_str += f"&{k}={v}"
        curl_command.append(post_data_str.lstrip("&"))

    method = kwargs.get("method", "")
    if method:
        curl_command.append("-X")
        curl_command.append(method)

    cookies = kwargs.get("cookies", "")
    if cookies:
        curl_command.append("-b")
        cookies_str = ""
        for k, v in cookies.items():
            cookies_str += f"{k}={v}; "
        curl_command.append(f'{cookies_str.rstrip(" ")}')

    path_override = kwargs.get("path_override", None)
    if path_override:
        curl_command.append("--request-target")
        curl_command.append(f"{path_override}")

    head_mode = kwargs.get("head_mode", None)
    if head_mode:
        curl_command.append("-I")

    raw_body = kwargs.get("raw_body", None)
    if raw_body:
        curl_command.append("-d")
        curl_command.append(raw_body)

    output_bytes = self.run(curl_command, text=False).stdout
    output = self.smart_decode(output_bytes)
    return output



# adopted from https://github.com/bugcrowd/HUNT

import re
from bbot.modules.base import BaseModule


hunt_param_dict = {
    "cmdi": ["daemon", "host", "upload", "dir", "execute", "download", "log", "ip", "cli", "cmd"],
    "debug": [
        "access",
        "admin",
        "dbg",
        "debug",
        "edit",
        "grant",
        "test",
        "alter",
        "clone",
        "create",
        "delete",
        "disable",
        "enable",
        "exec",
        "execute",
        "load",
        "make",
        "modify",
        "rename",
        "reset",
        "shell",
        "toggle",
        "adm",
        "root",
        "cfg",
        "config",
    ],
    "lfi": ["file", "document", "folder", "root", "path", "pg", "style", "pdf", "template", "php_path", "doc"],
    "idor": [
        "id",
        "user",
        "account",
        "number",
        "order",
        "no",
        "doc",
        "key",
        "email",
        "group",
        "profile",
        "edit",
        "report",
    ],
    "sqli": [
        "id",
        "select",
        "report",
        "role",
        "update",
        "query",
        "user",
        "name",
        "sort",
        "where",
        "search",
        "params",
        "category",
        "process",
        "row",
        "view",
        "table",
        "from",
        "sel",
        "results",
        "sleep",
        "fetch",
        "order",
        "keyword",
        "column",
        "field",
        "delete",
        "string",
        "number",
        "filter",
    ],
    "ssrf": [
        "dest",
        "redirect",
        "uri",
        "path",
        "continue",
        "url",
        "window",
        "next",
        "data",
        "reference",
        "site",
        "html",
        "val",
        "validate",
        "domain",
        "callback",
        "return",
        "page",
        "feed",
        "host",
        "port",
        "to",
        "out",
        "view",
        "dir",
        "show",
        "navigation",
        "open",
    ],
    "ssti": ["template", "preview", "id", "view", "activity", "name", "content", "redirect"],
}


class hunt(BaseModule):
    input_tag_regex = re.compile(r"<input.+?name=[\"\'](\w+)[\"\']")
    jquery_get_regex = re.compile(r"url:\s?[\"\'].+?\?(\w+)=")
    jquery_post_regex = re.compile(r"\$.post\([\'\"].+[\'\"].+\{(.+)\}")
    a_tag_regex = re.compile(r"<a\s+(?:[^>]*?\s+)?href=(?:[\"\'](.+\?.+?))[\"\'].+[>\s]")

    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING"]
    flags = ["active", "safe", "web-advanced"]
    meta = {"description": "Watch for commonly-exploitable HTTP parameters"}
    # accept all events regardless of scope distance
    scope_distance_modifier = None

    def extract_params(self, body):
        # check for input tags
        input_tag = self.input_tag_regex.findall(body)

        for i in input_tag:
            self.debug(f"FOUND PARAM ({i}) IN INPUT TAGS")
            yield i

        # check for jquery get parameters
        jquery_get = self.jquery_get_regex.findall(body)

        for i in jquery_get:
            self.debug(f"FOUND PARAM ({i}) IN JQUERY GET PARAMS")
            yield i

        # check for jquery post parameters
        jquery_post = self.jquery_post_regex.findall(body)
        if jquery_post:
            for i in jquery_post:
                for x in i.split(","):
                    s = x.split(":")[0].rstrip()
                    self.debug(f"FOUND PARAM ({s}) IN A JQUERY POST PARAMS")
                    yield s

        a_tag = self.a_tag_regex.findall(body)
        if a_tag:
            for url in a_tag:
                if url.startswith("http"):
                    url_parsed = self.helpers.parse_url(url)
                    if not self.scan.in_scope(url_parsed.netloc):
                        self.debug(f"Skipping checking for parameters because URL ({url}) is not in scope")
                        continue
                    i = url_parsed.query.split("&")
                else:
                    i = url.split("?")[1].split("&")
                for x in i:
                    s = x.split("=")[0]

                    self.debug(f"FOUND PARAM ({s}) IN A TAG GET PARAMS")
                    yield s

    def handle_event(self, event):
        body = event.data.get("body", "")
        for p in self.extract_params(body):
            for k in hunt_param_dict.keys():
                if p.lower() in hunt_param_dict[k]:
                    description = f"Found potential {k.upper()} parameter [{p}]"
                    data = {"host": str(event.host), "description": description}
                    url = event.data.get("url", "")
                    if url:
                        data["url"] = url
                    self.emit_event(data, "FINDING", event)



def test_threadpool():
    from concurrent.futures import ThreadPoolExecutor
    from bbot.core.helpers.threadpool import ThreadPoolWrapper, NamedLock, as_completed

    with ThreadPoolExecutor(max_workers=3) as executor:
        pool = ThreadPoolWrapper(executor)
        add_one = lambda x: x + 1
        futures = [pool.submit_task(add_one, y) for y in [0, 1, 2, 3, 4]]
        results = []
        for f in as_completed(futures):
            results.append(f.result())
        assert tuple(sorted(results)) == (1, 2, 3, 4, 5)

    nl = NamedLock(max_size=5)
    for i in range(50):
        nl.get_lock(str(i))
    assert len(nl._cache) == 5
    assert tuple(nl._cache.keys()) == tuple(hash(str(x)) for x in [45, 46, 47, 48, 49])



import json
import functools
import subprocess
from contextlib import suppress

from bbot.modules.base import BaseModule


class masscan(BaseModule):
    flags = ["active", "portscan", "aggressive"]
    watched_events = ["SCAN"]
    produced_events = ["OPEN_TCP_PORT"]
    meta = {"description": "Port scan IP subnets with masscan"}
    # 600 packets/s ~= entire private IP space in 8 hours
    options = {"ports": "80,443", "rate": 600, "wait": 10, "ping_first": False}
    options_desc = {
        "ports": "Ports to scan",
        "rate": "Rate in packets per second",
        "wait": "Seconds to wait for replies after scan is complete",
        "ping_first": "Only portscan hosts that reply to pings",
    }
    deps_ansible = [
        {
            "name": "install dev tools",
            "package": {"name": ["gcc", "git", "make"], "state": "present"},
            "become": True,
            "ignore_errors": True,
        },
        {
            "name": "Download masscan source code",
            "git": {
                "repo": "https://github.com/robertdavidgraham/masscan.git",
                "dest": "#{BBOT_TEMP}/masscan",
                "single_branch": True,
                "version": "master",
            },
        },
        {
            "name": "Build masscan",
            "command": {
                "chdir": "#{BBOT_TEMP}/masscan",
                "cmd": "make -j",
                "creates": "#{BBOT_TEMP}/masscan/bin/masscan",
            },
        },
        {
            "name": "Install masscan",
            "copy": {"src": "#{BBOT_TEMP}/masscan/bin/masscan", "dest": "#{BBOT_TOOLS}/", "mode": "u+x,g+x,o+x"},
        },
    ]
    _qsize = 100

    def setup(self):
        self.ports = self.config.get("ports", "80,443")
        self.rate = self.config.get("rate", 600)
        self.wait = self.config.get("wait", 10)
        self.ping_first = self.config.get("ping_first", False)
        self.alive_hosts = dict()
        # make a quick dry run to validate ports etc.
        self._target_findkey = "9.8.7.6"
        try:
            dry_run_command = self._build_masscan_command(self._target_findkey, dry_run=True)
            dry_run_result = self.helpers.run(dry_run_command)
            self.masscan_config = dry_run_result.stdout
            self.masscan_config = "\n".join(l for l in self.masscan_config.splitlines() if "nocapture" not in l)
        except subprocess.CalledProcessError as e:
            self.warning(f"Error in masscan: {e.stderr}")
            return False
        self.helpers.depsinstaller.ensure_root(message="Masscan requires root privileges")
        return True

    def handle_event(self, event):
        exclude, invalid_exclude = self._build_targets(self.scan.blacklist)
        targets, invalid_targets = self._build_targets(self.scan.whitelist)
        if invalid_exclude > 0:
            self.warning(
                f"Masscan can only accept IP addresses or IP ranges for blacklist ({invalid_exclude:,} blacklisted were hostnames)"
            )
        if invalid_targets > 0:
            self.warning(
                f"Masscan can only accept IP addresses or IP ranges as target ({invalid_targets:,} targets were hostnames)"
            )

        if not targets:
            self.warning("No targets specified")
            return

        # ping scan
        if self.ping_first:
            self.verbose("Starting masscan (ping scan)")

            def append_alive_host(host, source):
                host_event = self.make_event(host, "IP_ADDRESS", source=self.scan.whitelist.get(host))
                self.alive_hosts[host] = host_event
                self.emit_event(host_event)

            self.masscan(targets, result_callback=append_alive_host, exclude=exclude, ping=True)
            targets = ",".join(str(h) for h in self.alive_hosts)
            if not targets:
                self.warning("No hosts responded to pings")
                return

        # TCP SYN scan
        self.verbose("Starting masscan (TCP SYN scan)")
        self.masscan(targets, result_callback=self.emit_open_tcp_port, exclude=exclude, event=event)
        # save memory
        self.alive_hosts.clear()

    def masscan(self, targets, result_callback, exclude=None, event=None, ping=False):
        # config file
        masscan_config = self.masscan_config.replace(self._target_findkey, targets)
        self.debug("Masscan config:")
        for line in masscan_config.splitlines():
            self.debug(line)
        config_file = self.helpers.tempfile(masscan_config)
        # output file
        process_output = functools.partial(self.process_output, source=event, result_callback=result_callback)
        json_output_file = self.helpers.tempfile_tail(process_output)
        # command
        command = self._build_masscan_command(config=config_file, exclude=exclude, ping=ping)
        command += ("-oJ", json_output_file)
        # execute
        self.helpers.run(command, sudo=True)

    def _build_masscan_command(self, targets=None, config=None, exclude=None, dry_run=False, ping=False):
        command = ("masscan", "--rate", self.rate, "--wait", self.wait, "--open-only")
        if targets is not None:
            command += (targets,)
        if config is not None:
            command += ("-c", config)
        if ping:
            command += ("--ping",)
        elif not dry_run:
            command += ("-p", self.ports)
        if exclude is not None:
            command += ("--exclude", exclude)
        if dry_run:
            command += ("--echo",)
        return command

    def process_output(self, line, source, result_callback):
        try:
            j = json.loads(line)
        except Exception:
            return
        ip = j.get("ip", "")
        if not ip:
            return
        ports = j.get("ports", [])
        if not ports:
            return
        for p in ports:
            proto = p.get("proto", "")
            port_number = p.get("port", "")
            if proto == "" or port_number == "":
                continue
            result = str(ip)
            if proto != "icmp":
                result += f":{port_number}"
            with suppress(KeyError):
                source = self.alive_hosts[ip]
            result_callback(result, source=source)

    def emit_open_tcp_port(self, data, source):
        self.emit_event(data, "OPEN_TCP_PORT", source=source)

    def _build_targets(self, target):
        invalid_targets = 0
        targets = []
        for t in target:
            t = self.helpers.make_ip_type(t.data)
            if isinstance(t, str):
                invalid_targets += 1
            else:
                targets.append(t)
        return ",".join(str(t) for t in targets), invalid_targets



import sys
import argparse
from pathlib import Path
from omegaconf import OmegaConf
from contextlib import suppress

from ..errors import ArgumentError
from ...modules import module_loader
from ..helpers.misc import chain_lists
from ..helpers.logger import log_to_stderr

module_choices = sorted(set(module_loader.configs(type="scan")))
output_module_choices = sorted(set(module_loader.configs(type="output")))

flag_choices = set()
for m, c in module_loader.preloaded().items():
    flag_choices.update(set(c.get("flags", [])))


class BBOTArgumentParser(argparse.ArgumentParser):
    _dummy = False

    def parse_args(self, *args, **kwargs):
        """
        Allow space or comma-separated entries for modules and targets
        For targets, also allow input files containing additional targets
        """
        ret = super().parse_args(*args, **kwargs)
        ret.modules = chain_lists(ret.modules)
        ret.output_modules = chain_lists(ret.output_modules)
        ret.targets = chain_lists(ret.targets, try_files=True, msg="Reading targets from file: {filename}")
        ret.whitelist = chain_lists(ret.whitelist, try_files=True, msg="Reading whitelist from file: {filename}")
        ret.blacklist = chain_lists(ret.blacklist, try_files=True, msg="Reading blacklist from file: {filename}")
        ret.flags = chain_lists(ret.flags)
        ret.require_flags = chain_lists(ret.require_flags)
        for m in ret.modules:
            if m not in module_choices and not self._dummy:
                raise ArgumentError(f'Module "{m}" is not valid. Choose from: {",".join(module_choices)}')
        for m in ret.exclude_modules:
            if m not in module_choices and not self._dummy:
                raise ArgumentError(f'Cannot exclude module "{m}". Choose from: {",".join(module_choices)}')
        for m in ret.output_modules:
            if m not in output_module_choices and not self._dummy:
                raise ArgumentError(
                    f'Output module "{m}" is not valid. Choose from: {",".join(output_module_choices)}'
                )
        for f in set(ret.flags + ret.require_flags):
            if f not in flag_choices and not self._dummy:
                raise ArgumentError(f'Flag "{f}" is not valid. Choose from: {",".join(sorted(flag_choices))}')
        return ret


class DummyArgumentParser(BBOTArgumentParser):
    _dummy = True

    def error(self, message):
        pass


epilog = """EXAMPLES

    list modules:
        bbot -l

    subdomain enumeration:
        bbot -t evilcorp.com -f subdomain-enum -m httpx

    passive modules only:
        bbot -t evilcorp.com -f passive

    subdomains + web screenshots:
        bbot -t targets.txt -f subdomain-enum -m httpx gowitness --name my_scan --output-dir .

    subdomains + basic web scanning:
        bbot -t evilcorp.com -f subdomain-enum web-basic

    single module:
        bbot -t evilcorp.com -m github -c modules.github.api_key=deadbeef

    web spider + advanced web scan:
        bbot -t www.evilcorp.com -m httpx -f web-basic web-advanced -c web_spider_distance=2 web_spider_depth=2

    subdomains + emails + cloud buckets + portscan + screenshots + nuclei:
        bbot -t evilcorp.com -f subdomain-enum email-enum cloud-enum web-basic -m naabu gowitness nuclei --allow-deadly

"""


parser = BBOTArgumentParser(
    description="Bighuge BLS OSINT Tool", formatter_class=argparse.RawTextHelpFormatter, epilog=epilog
)
dummy_parser = DummyArgumentParser(
    description="Bighuge BLS OSINT Tool", formatter_class=argparse.RawTextHelpFormatter, epilog=epilog
)
for p in (parser, dummy_parser):
    p.add_argument("--help-all", action="store_true", help="Display full help including module config options")
    target = p.add_argument_group(title="Target")
    target.add_argument("-t", "--targets", nargs="+", default=[], help="Targets to seed the scan", metavar="TARGET")
    target.add_argument(
        "-w",
        "--whitelist",
        nargs="+",
        default=[],
        help="What's considered in-scope (by default it's the same as --targets)",
    )
    target.add_argument("-b", "--blacklist", nargs="+", default=[], help="Don't touch these things")
    target.add_argument(
        "--strict-scope",
        action="store_true",
        help="Don't consider subdomains of target/whitelist to be in-scope",
    )
    p.add_argument("-n", "--name", help="Name of scan (default: random)", metavar="SCAN_NAME")
    p.add_argument(
        "-m",
        "--modules",
        nargs="+",
        default=[],
        help=f'Modules to enable. Choices: {",".join(module_choices)}',
        metavar="MODULE",
    )
    p.add_argument("-l", "--list-modules", action="store_true", help=f"List available modules.")
    p.add_argument("-em", "--exclude-modules", nargs="+", default=[], help=f"Exclude these modules.", metavar="MODULE")
    p.add_argument(
        "-f",
        "--flags",
        nargs="+",
        default=[],
        help=f'Enable modules by flag. Choices: {",".join(sorted(flag_choices))}',
        metavar="FLAG",
    )
    p.add_argument(
        "-rf",
        "--require-flags",
        nargs="+",
        default=[],
        help=f"Disable modules that don't have these flags (e.g. --require-flags passive)",
        metavar="FLAG",
    )
    p.add_argument(
        "-ef",
        "--exclude-flags",
        nargs="+",
        default=[],
        help=f"Disable modules with these flags. (e.g. --exclude-flags brute-force)",
        metavar="FLAG",
    )
    p.add_argument(
        "-om",
        "--output-modules",
        nargs="+",
        default=["human", "json", "csv"],
        help=f'Output module(s). Choices: {",".join(output_module_choices)}',
        metavar="MODULE",
    )
    p.add_argument(
        "-o",
        "--output-dir",
        metavar="DIR",
    )
    p.add_argument(
        "-c",
        "--config",
        nargs="*",
        help="custom config file, or configuration options in key=value format: 'modules.shodan.api_key=1234'",
        metavar="CONFIG",
    )
    p.add_argument("--allow-deadly", action="store_true", help="Enable the use of highly aggressive modules")
    p.add_argument("-v", "--verbose", action="store_true", help="Be more verbose")
    p.add_argument("-d", "--debug", action="store_true", help="Enable debugging")
    p.add_argument("-s", "--silent", action="store_true", help="Be quiet")
    p.add_argument("--force", action="store_true", help="Run scan even if module setups fail")
    p.add_argument("-y", "--yes", action="store_true", help="Skip scan confirmation prompt")
    p.add_argument("--dry-run", action="store_true", help=f"Abort before executing scan")
    p.add_argument(
        "--current-config",
        action="store_true",
        help="Show current config in YAML format",
    )
    wordcloud = p.add_argument_group(
        title="Word cloud", description="Save/load wordlist of common words gathered during a scan"
    )
    wordcloud.add_argument(
        "--save-wordcloud", help="Output wordcloud to custom file when the scan completes", metavar="FILE"
    )
    wordcloud.add_argument("--load-wordcloud", help="Load wordcloud from a custom file", metavar="FILE")
    deps = p.add_argument_group(
        title="Module dependencies", description="Control how modules install their dependencies"
    )
    g2 = deps.add_mutually_exclusive_group()
    g2.add_argument("--no-deps", action="store_true", help="Don't install module dependencies")
    g2.add_argument("--force-deps", action="store_true", help="Force install all module dependencies")
    g2.add_argument("--retry-deps", action="store_true", help="Try again to install failed module dependencies")
    g2.add_argument(
        "--ignore-failed-deps", action="store_true", help="Run modules even if they have failed dependencies"
    )
    g2.add_argument("--install-all-deps", action="store_true", help="Install dependencies for all modules")
    agent = p.add_argument_group(title="Agent", description="Report back to a central server")
    agent.add_argument("-a", "--agent-mode", action="store_true", help="Start in agent mode")
    misc = p.add_argument_group(title="Misc")
    misc.add_argument("--version", action="store_true", help="show BBOT version and exit")


cli_options = None
with suppress(Exception):
    cli_options = dummy_parser.parse_args()


def get_config():
    cli_config = []
    with suppress(Exception):
        if cli_options.config:
            cli_config = cli_options.config
    if len(cli_config) == 1:
        filename = Path(cli_config[0]).resolve()
        if filename.is_file():
            try:
                conf = OmegaConf.load(str(filename))
                log_to_stderr(f"Loaded custom config from {filename}")
                return conf
            except Exception as e:
                log_to_stderr(f"Error parsing custom config at {filename}: {e}", level="ERROR")
                sys.exit(2)
    try:
        return OmegaConf.from_cli(cli_config)
    except Exception as e:
        log_to_stderr(f"Error parsing command-line config: {e}", level="ERROR")
        sys.exit(2)



from bbot.modules.crobat import crobat


class certspotter(crobat):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query Certspotter's API for subdomains"}

    base_url = "https://api.certspotter.com/v1"

    def request_url(self, query):
        url = f"{self.base_url}/issuances?domain={self.helpers.quote(query)}&include_subdomains=true&expand=dns_names"
        return self.helpers.request(url)

    def parse_results(self, r, query):
        json = r.json()
        if json:
            for r in json:
                for dns_name in r.get("dns_names", []):
                    yield dns_name.lstrip(".*").rstrip(".")



from bbot.modules.shodan_dns import shodan_dns


class virustotal(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query VirusTotal's API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "VirusTotal API Key"}

    base_url = "https://www.virustotal.com/api/v3"

    def setup(self):
        self.api_key = self.config.get("api_key", "")
        self.headers = {"x-apikey": self.api_key}
        return super().setup()

    def ping(self):
        # virustotal does not have a ping function
        return

    def request_url(self, query):
        url = f"{self.base_url}/domains/{self.helpers.quote(query)}/subdomains"
        return self.helpers.request(url, headers=self.headers)

    def parse_results(self, r, query):
        results = set()
        text = getattr(r, "text", "")
        for match in self.helpers.regexes.dns_name_regex.findall(text):
            match = match.lower()
            if match.endswith(query):
                results.add(match)
        return results



import re

from bbot.modules.base import BaseModule

valid_chars = "ETAONRISHDLFCMUGYPWBVKJXQZ0123456789_-$~()&!#%'@^`{}]]"


def encode_all(string):
    return "".join("%{0:0>2}".format(format(ord(char), "x")) for char in string)


class iis_shortnames(BaseModule):
    watched_events = ["URL"]
    produced_events = ["URL_HINT"]
    flags = ["active", "safe", "web-basic", "iis-shortnames"]
    meta = {"description": "Check for IIS shortname vulnerability"}
    options = {"detect_only": True}
    options_desc = {"detect_only": "Only detect the vulnerability and do not run the shortname scanner"}
    in_scope_only = True

    max_event_handlers = 8

    def detect(self, target):
        technique = None
        headers = {}
        detected = []
        random_string = self.helpers.rand_string(8)
        control_url = f"{target}{random_string}*~1*/a.aspx"
        test_url = f"{target}*~1*/a.aspx"

        for method in ["GET", "POST", "OPTIONS", "DEBUG", "HEAD", "TRACE"]:
            control = self.helpers.request(method=method, headers=headers, url=control_url)
            test = self.helpers.request(method=method, headers=headers, url=test_url)
            if (control != None) and (test != None):
                if (control.status_code != 404) and (test.status_code == 404):
                    detected.append(method)
                    technique = "400/404 HTTP Code"

                elif ("Error Code</th><td>0x80070002" in control.text) and (
                    "Error Code</th><td>0x00000000" in test.text
                ):
                    detected.append(method)
                    technique = "HTTP Body Error Message"
        return detected, technique

    def duplicate_check(self, target, method, url_hint):
        duplicates = []
        headers = {}
        count = 2
        base_hint = re.sub(r"~\d", "", url_hint)
        suffix = "\\a.aspx"

        while 1:
            payload = encode_all(f"{base_hint}~{str(count)}*")
            url = f"{target}{payload}{suffix}"

            duplicate_check_results = self.helpers.request(method=method, headers=headers, url=url)
            if duplicate_check_results.status_code != 404:
                break
            else:
                duplicates.append(f"{base_hint}~{str(count)}")
                count += 1

            if count > 5:
                self.warning("Found more than 5 files with the same shortname. Will stop further duplicate checking.")
                break

        return duplicates

    def threaded_request(self, method, url):
        r = self.helpers.request(method=method, url=url)
        if r is not None:
            if r.status_code == 404:
                return True

    def solve_shortname_recursive(self, method, target, prefix, extension_mode=False):
        url_hint_list = []
        found_results = False

        futures = {}
        for c in valid_chars:
            suffix = "\\a.aspx"
            wildcard = "*" if extension_mode else "*~1*"
            payload = encode_all(f"{prefix}{c}{wildcard}")
            url = f"{target}{payload}{suffix}"
            future = self.submit_task(self.threaded_request, method, url)
            futures[future] = c

        for future in self.helpers.as_completed(futures):
            c = futures[future]
            result = future.result()
            if result:
                found_results = True

                # check to make sure the file isn't shorter than 6 characters
                wildcard = "~1*"
                payload = encode_all(f"{prefix}{c}{wildcard}")
                url = f"{target}{payload}{suffix}"
                r = self.helpers.request(method=method, url=url)
                if r is not None:
                    if r.status_code == 404:
                        url_hint_list.append(f"{prefix}{c}")

                url_hint_list += self.solve_shortname_recursive(method, target, f"{prefix}{c}", extension_mode)
        if len(prefix) > 0 and found_results == False:
            url_hint_list.append(f"{prefix}")
        return url_hint_list

    def handle_event(self, event):
        normalized_url = event.data.rstrip("/") + "/"
        vulnerable_methods, technique = self.detect(normalized_url)
        if vulnerable_methods:
            description = f"IIS Shortname Vulnerability Detected. Potentially Vulnerable methods: [{','.join(vulnerable_methods)}] Technique: [{technique}]"
            self.emit_event(
                {"severity": "LOW", "host": str(event.host), "url": normalized_url, "description": description},
                "VULNERABILITY",
                event,
            )
            if not self.config.get("detect_only"):
                valid_method_confirmed = False
                for m in vulnerable_methods:
                    if valid_method_confirmed:
                        break

                    file_name_hints = self.solve_shortname_recursive(m, normalized_url, "")
                    if len(file_name_hints) == 0:
                        continue
                    else:
                        valid_method_confirmed = True

                    file_name_hints = [f"{x}~1" for x in file_name_hints]
                    url_hint_list = []

                    file_name_hints_dedupe = file_name_hints[:]

                    for x in file_name_hints_dedupe:
                        duplicates = self.duplicate_check(normalized_url, m, x)
                        if duplicates:
                            file_name_hints += duplicates

                    for y in file_name_hints:
                        file_name_extension_hints = self.solve_shortname_recursive(
                            m, normalized_url, f"{y}.", extension_mode=True
                        )
                        for z in file_name_extension_hints:
                            url_hint_list.append(z)

                    for url_hint in url_hint_list:
                        if url_hint.endswith("."):
                            url_hint = url_hint.rstrip(".")
                        if "." in url_hint:
                            hint_type = "shortname-file"
                        else:
                            hint_type = "shortname-directory"
                        self.emit_event(f"{normalized_url}/{url_hint}", "URL_HINT", event, tags=[hint_type])



import logging
import ipaddress

from bbot.core.helpers import regexes
from bbot.core.helpers.url import clean_url
from bbot.core.helpers.punycode import smart_decode_punycode
from bbot.core.helpers.misc import split_host_port, make_netloc

log = logging.getLogger("bbot.core.helpers.")


def validator(func):
    """
    Decorator for squashing all errors into ValueError
    """

    def validate_wrapper(*args, **kwargs):
        try:
            return func(*args)
        except Exception as e:
            raise ValueError(f"Validation failed for {args}, {kwargs}: {e}")

    return validate_wrapper


@validator
def validate_port(port):
    return max(1, min(65535, int(str(port))))


@validator
def validate_open_port(open_port):
    host, port = split_host_port(open_port)
    port = validate_port(port)
    host = validate_host(host)
    if host and port:
        return make_netloc(host, port)


@validator
def validate_host(host):
    # stringify, strip and lowercase
    host = str(host).strip().lower()
    # handle IPv6 netlocs
    if host.startswith("["):
        host = host.split("[")[-1].split("]")[0]
    try:
        # try IPv6 first
        ip = ipaddress.IPv6Address(host)
        return str(ip)
    except Exception:
        # if IPv6 fails, strip ports and root zone
        host = host.split(":")[0].rstrip(".")
        try:
            ip = ipaddress.IPv4Address(host)
            return str(ip)
        except Exception:
            # finally, try DNS_NAME
            host = smart_decode_punycode(host.lstrip("*."))
            for r in regexes.event_type_regexes["DNS_NAME"]:
                if r.match(host):
                    return host
            if regexes._hostname_regex.match(host):
                return host
    assert False, f'Invalid hostname: "{host}"'


@validator
def validate_url(url):
    return validate_url_parsed(url).geturl()


@validator
def validate_url_parsed(url):
    url = str(url).strip()
    if not any(r.match(url) for r in regexes.event_type_regexes["URL"]):
        assert False, f'Invalid URL: "{url}"'
    return clean_url(url)


@validator
def validate_severity(severity):
    severity = str(severity).strip().upper()
    if not severity in ("INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"):
        raise ValueError(f"Invalid severity: {severity}")
    return severity


@validator
def validate_email(email):
    email = smart_decode_punycode(str(email).strip().lower())
    if any(r.match(email) for r in regexes.event_type_regexes["EMAIL_ADDRESS"]):
        return email
    assert False, f'Invalid email: "{email}"'


def soft_validate(s, t):
    """
    Friendly validation wrapper that returns True/False instead of raising an error

    is_valid_url = soft_validate("http://evilcorp.com", "url")
    is_valid_host = soft_validate("http://evilcorp.com", "host")
    """
    try:
        validator_fn = globals()[f"validate_{t.strip().lower()}"]
    except KeyError:
        raise ValueError(f'No validator for type "{t}"')
    try:
        validator_fn(s)
        return True
    except ValueError:
        return False



import random
import ipaddress

from bbot.modules.internal.base import BaseInternalModule


class speculate(BaseInternalModule):
    """
    Bridge the gap between ranges and ips, or ips and open ports
    in situations where e.g. a port scanner isn't enabled
    """

    watched_events = ["IP_RANGE", "URL", "URL_UNVERIFIED", "DNS_NAME", "IP_ADDRESS", "HTTP_RESPONSE", "STORAGE_BUCKET"]
    produced_events = ["DNS_NAME", "OPEN_TCP_PORT", "IP_ADDRESS", "FINDING"]
    flags = ["passive"]
    meta = {"description": "Derive certain event types from others by common sense"}

    options = {"max_hosts": 65536}
    options_desc = {"max_hosts": "Max number of IP_RANGE hosts to convert into IP_ADDRESS events"}
    max_event_handlers = 5
    scope_distance_modifier = 0
    _scope_shepherding = False
    _priority = 4

    def setup(self):
        self.open_port_consumers = any(["OPEN_TCP_PORT" in m.watched_events for m in self.scan.modules.values()])
        self.portscanner_enabled = any(["portscan" in m.flags for m in self.scan.modules.values()])
        self.range_to_ip = True
        target_len = len(self.scan.target)
        if target_len > self.config.get("max_hosts", 65536):
            if not self.portscanner_enabled:
                self.hugewarning(
                    f"Selected target ({target_len:,} hosts) is too large, skipping IP_RANGE --> IP_ADDRESS speculation"
                )
                self.hugewarning(f"Enabling a port scanner module is highly recommended")
            self.range_to_ip = False
        return True

    def handle_event(self, event):
        # generate individual IP addresses from IP range
        if event.type == "IP_RANGE" and self.range_to_ip:
            net = ipaddress.ip_network(event.data)
            ips = list(net)
            random.shuffle(ips)
            for ip in ips:
                self.emit_event(ip, "IP_ADDRESS", source=event, internal=True)

        # parent domains
        if event.type == "DNS_NAME":
            parent = self.helpers.parent_domain(event.data)
            if parent != event.data:
                self.emit_event(parent, "DNS_NAME", source=event, internal=True)

        # generate open ports
        emit_open_ports = self.open_port_consumers and not self.portscanner_enabled
        # from URLs
        if event.type == "URL" or (event.type == "URL_UNVERIFIED" and emit_open_ports):
            if event.host and event.port not in (80, 443):
                self.emit_event(
                    self.helpers.make_netloc(event.host, event.port), "OPEN_TCP_PORT", source=event, internal=True
                )
        # from hosts
        if emit_open_ports:
            # don't act on unresolved DNS_NAMEs

            usable_dns = False
            if event.type == "DNS_NAME":
                if "a-record" in event.tags or "aaaa-record" in event.tags:
                    usable_dns = True

            if event.type == "IP_ADDRESS" or usable_dns:
                self.emit_event(self.helpers.make_netloc(event.data, 80), "OPEN_TCP_PORT", source=event, internal=True)
                self.emit_event(
                    self.helpers.make_netloc(event.data, 443), "OPEN_TCP_PORT", source=event, internal=True
                )

    def filter_event(self, event):
        # don't accept IP_RANGE --> IP_ADDRESS events from self
        if str(event.module) == "speculate":
            if not (event.type == "IP_ADDRESS" and str(getattr(event.source, "type")) == "IP_RANGE"):
                return False
        # don't accept errored DNS_NAMEs
        if any(t in event.tags for t in ("dns-error", "unresolved")):
            return False
        return True



from bbot.modules.base import BaseModule
from bbot.core.errors import InteractshError


ssrf_params = [
    "Dest",
    "Redirect",
    "URI",
    "Path",
    "Continue",
    "URL",
    "Window",
    "Next",
    "Data",
    "Reference",
    "Site",
    "HTML",
    "Val",
    "Validate",
    "Domain",
    "Callback",
    "Return",
    "Page",
    "Feed",
    "Host",
    "Port",
    "To",
    "Out",
    "View",
    "Dir",
    "Show",
    "Navigation",
    "Open",
]


class BaseSubmodule:
    technique_description = "base technique description"
    severity = "INFO"
    paths = None

    def __init__(self, parent_module):
        self.parent_module = parent_module
        self.test_paths = self.create_paths()

    def set_base_url(self, event):
        return f"{event.parsed.scheme}://{event.parsed.netloc}"

    def create_paths(self):
        return self.paths

    def test(self, event):
        base_url = self.set_base_url(event)

        for test_path in self.test_paths:
            subdomain_tag = self.parent_module.helpers.rand_string(4)
            test_path_prepared = test_path.replace(
                "SSRF_CANARY", f"{subdomain_tag}.{self.parent_module.interactsh_domain}"
            )
            test_url = f"{base_url}{test_path_prepared}"
            self.parent_module.debug(f"Sending request to URL: {test_url}")
            r = self.parent_module.helpers.curl(url=test_url)
            if r:
                self.process(event, r, subdomain_tag)

    def process(self, event, r, subdomain_tag):
        response_token = self.parent_module.interactsh_domain.split(".")[0][::-1]
        if response_token in r:
            read_response = True
        else:
            read_response = False

        self.parent_module.interactsh_subdomain_tags[subdomain_tag] = (
            event,
            self.technique_description,
            self.severity,
            read_response,
        )


class Generic_SSRF(BaseSubmodule):
    technique_description = "Generic SSRF (GET)"
    severity = "HIGH"

    def set_base_url(self, event):
        return event.data

    def create_paths(self):
        query_string = ""
        for param in ssrf_params:
            query_string += f"{param}=http://SSRF_CANARY&"

        query_string_lower = ""
        for param in ssrf_params:
            query_string_lower += f"{param.lower()}=http://SSRF_CANARY&"

        return [f"?{query_string.rstrip('&')}", f"?{query_string_lower.rstrip('&')}"]


class Generic_SSRF_POST(BaseSubmodule):
    technique_description = "Generic SSRF (POST)"
    severity = "HIGH"

    def set_base_url(self, event):
        return event.data

    def test(self, event):
        test_url = f"{event.data}"

        subdomain_tag = self.parent_module.helpers.rand_string(4, digits=False)
        post_data = {}
        for param in ssrf_params:
            post_data[param] = f"http://{subdomain_tag}.{self.parent_module.interactsh_domain}"

        subdomain_tag_lower = self.parent_module.helpers.rand_string(4, digits=False)
        post_data_lower = {
            k.lower(): f"http://{subdomain_tag_lower}.{self.parent_module.interactsh_domain}"
            for k, v in post_data.items()
        }

        post_data_list = [(subdomain_tag, post_data), (subdomain_tag_lower, post_data_lower)]

        for tag, pd in post_data_list:
            r = self.parent_module.helpers.curl(url=test_url, method="POST", post_data=pd)
            self.process(event, r, tag)


class Generic_XXE(BaseSubmodule):
    technique_description = "Generic XXE"
    severity = "HIGH"
    paths = None

    def test(self, event):
        rand_entity = self.parent_module.helpers.rand_string(4, digits=False)
        subdomain_tag = self.parent_module.helpers.rand_string(4, digits=False)

        post_body = f"""<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE foo [
<!ELEMENT foo ANY >
<!ENTITY % {rand_entity} SYSTEM "http://{subdomain_tag}.{self.parent_module.interactsh_domain}" >
]>
<foo>&{rand_entity};</foo>"""
        test_url = f"{event.parsed.scheme}://{event.parsed.netloc}/"
        r = self.parent_module.helpers.curl(
            url=test_url, method="POST", raw_body=post_body, headers={"Content-type": "application/xml"}
        )
        if r:
            self.process(event, r, subdomain_tag)


class generic_ssrf(BaseModule):
    watched_events = ["URL"]
    produced_events = ["VULNERABILITY"]
    flags = ["active", "aggressive", "web-advanced"]
    meta = {"description": "Check for generic SSRFs"}
    in_scope_only = True

    deps_apt = ["curl"]

    def setup(self):
        self.submodules = {}
        self.interactsh_subdomain_tags = {}
        self.severity = None
        self.generic_only = self.config.get("generic_only", False)

        if self.scan.config.get("interactsh_disable", False) == False:
            try:
                self.interactsh_instance = self.helpers.interactsh()
                self.interactsh_domain = self.interactsh_instance.register(callback=self.interactsh_callback)
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")
                return False
        else:
            self.warning(
                "The generic_ssrf module is completely dependent on interactsh to function, but it is disabled globally. Aborting."
            )
            return None

        # instantiate submodules
        for m in BaseSubmodule.__subclasses__():
            if m.__name__.startswith("Generic_"):
                self.verbose(f"Starting generic_ssrf submodule: {m.__name__}")
                self.submodules[m.__name__] = m(self)

        return True

    def handle_event(self, event):
        self.test_submodules(self.submodules, event)

    def test_submodules(self, submodules, event, **kwargs):
        for s in submodules.values():
            s.test(event, **kwargs)

    def interactsh_callback(self, r):
        full_id = r.get("full-id", None)
        if full_id:
            if "." in full_id:
                match = self.interactsh_subdomain_tags.get(full_id.split(".")[0])
                if not match:
                    return
                matched_event = match[0]
                matched_technique = match[1]
                matched_severity = match[2]
                matched_read_response = str(match[3])

                self.emit_event(
                    {
                        "severity": matched_severity,
                        "host": str(matched_event.host),
                        "url": matched_event.data,
                        "description": f"Out-of-band interaction: [{matched_technique}] [{r.get('protocol').upper()}] Read Response: {matched_read_response}",
                    },
                    "VULNERABILITY",
                    matched_event,
                )
            else:
                # this is likely caused by something trying to resolve the base domain first and can be ignored
                self.debug("skipping result because subdomain tag was missing")

    def finish(self):
        from time import sleep

        sleep(5)

        try:
            for r in self.interactsh_instance.poll():
                self.interactsh_callback(r)
        except InteractshError as e:
            self.debug(f"Error in interact.sh: {e}")



from .crobat import crobat


class leakix(crobat):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query leakix.net for subdomains"}

    base_url = "https://leakix.net"

    def handle_event(self, event):
        query = self.make_query(event)
        headers = {"Accept": "application/json"}
        r = self.helpers.request(f"{self.base_url}/domain/{self.helpers.quote(query)}", headers=headers)
        if not r:
            return
        try:
            j = r.json()
        except Exception:
            self.verbose(f"Error decoding JSON")
            return
        services = j.get("Services", [])
        if services:
            for s in services:
                if s.get("event_type", "") != "service":
                    continue
                host = s.get("host", "")
                if not host:
                    continue
                if host == event:
                    source_event = event
                else:
                    source_event = self.make_event(host, "DNS_NAME", source=event)
                    if not source_event:
                        continue
                    self.emit_event(source_event)
                ssl = s.get("ssl", {})
                if not ssl:
                    continue
                certificate = ssl.get("certificate", {})
                if not certificate:
                    continue
                cert_domains = set()
                cn = self.clean_dns_name(certificate.get("cn", ""))
                if cn:
                    cert_domains.add(cn)
                domains = certificate.get("domain", [])
                if domains:
                    for d in domains:
                        d = self.clean_dns_name(d)
                        if d:
                            cert_domains.add(d)
                for d in cert_domains:
                    if d != host:
                        self.emit_event(d, "DNS_NAME", source=source_event)

    @staticmethod
    def clean_dns_name(dns_name):
        return str(dns_name).strip().lower().lstrip(".*")



from .shodan_dns import shodan_dns


class securitytrails(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the SecurityTrails API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "SecurityTrails API key"}

    base_url = "https://api.securitytrails.com/v1"

    def setup(self):
        self.limit = 100
        return super().setup()

    def ping(self):
        r = self.helpers.request(f"{self.base_url}/ping?apikey={self.api_key}")
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    def request_url(self, query):
        url = f"{self.base_url}/domain/{query}/subdomains?apikey={self.api_key}"
        return self.helpers.request(url)

    def parse_results(self, r, query):
        j = r.json()
        if isinstance(j, dict):
            for host in j.get("subdomains", []):
                yield f"{host}.{query}"



from .crobat import crobat


class shodan_dns(crobat):
    """
    A typical module for authenticated, API-based subdomain enumeration
    Inherited by several other modules including securitytrails, c99.nl, etc.
    """

    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query Shodan for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "Shodan API key"}

    base_url = "https://api.shodan.io"

    def setup(self):
        super().setup()
        self.api_key = self.config.get("api_key", "")
        if self.auth_secret:
            try:
                self.ping()
                self.hugesuccess(f"API is ready")
                return True
            except Exception as e:
                return None, f"Error with API ({str(e).strip()})"
        else:
            return None, "No API key set"

    def ping(self):
        r = self.helpers.request(f"{self.base_url}/api-info?key={self.api_key}")
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    def request_url(self, query):
        url = f"{self.base_url}/dns/domain/{self.helpers.quote(query)}?key={self.api_key}"
        return self.helpers.request(url)

    def parse_results(self, r, query):
        json = r.json()
        if json:
            for hostname in json.get("subdomains"):
                yield f"{hostname}.{query}"

    @property
    def auth_secret(self):
        return self.api_key



from ..bbot_fixtures import *  # noqa: F401


def test_manager(bbot_config, bbot_scanner):
    dns_config = OmegaConf.merge(default_config, OmegaConf.create({"dns_resolution": True}))

    # test _emit_event
    results = []
    success_callback = lambda e: results.append("success")
    scan1 = bbot_scanner("127.0.0.1", modules=["ipneighbor"], config=dns_config)
    scan1.status = "RUNNING"
    manager = scan1.manager
    manager.distribute_event = lambda e: results.append(e)
    localhost = scan1.make_event("127.0.0.1", source=scan1.root_event)

    class DummyModule1:
        _type = "output"
        suppress_dupes = True

    class DummyModule2:
        _type = "DNS"
        suppress_dupes = True

    localhost.module = DummyModule1()
    # test abort_if
    manager._emit_event(localhost, abort_if=lambda e: e.module._type == "output")
    assert len(results) == 0
    manager.events_accepted.clear()
    manager._emit_event(
        localhost, on_success_callback=success_callback, abort_if=lambda e: e.module._type == "plumbus"
    )
    assert localhost in results
    assert "success" in results
    results.clear()
    # test deduplication
    manager._emit_event(localhost, on_success_callback=success_callback)
    assert len(results) == 0
    # test dns resolution
    googledns = scan1.make_event("8.8.8.8", source=scan1.root_event)
    googledns.module = DummyModule2()
    googledns.source = "asdf"
    googledns.make_in_scope()
    event_children = []
    manager.emit_event = lambda e, *args, **kwargs: event_children.append(e)
    manager._emit_event(googledns)
    assert len(event_children) > 0
    assert googledns in results
    results.clear()
    event_children.clear()
    # same dns event
    manager._emit_event(googledns)
    assert len(results) == 0
    assert len(event_children) == 0
    # same dns event but with _force_output
    googledns._force_output = True
    manager._emit_event(googledns)
    assert googledns in results
    assert len(event_children) == 0
    googledns._force_output = False
    results.clear()
    # same dns event but different source
    source_event = manager.scan.make_event("1.2.3.4", "IP_ADDRESS", source=manager.scan.root_event)
    source_event._resolved.set()
    googledns.source = source_event
    manager._emit_event(googledns)
    assert len(event_children) == 0
    assert googledns in results

    # event filtering based on scope_distance
    output_queue = []
    module_queue = []
    scan1 = bbot_scanner("127.0.0.1", modules=["ipneighbor"], output_modules=["json"], config=bbot_config)
    scan1.status = "RUNNING"
    scan1.load_modules()
    manager = scan1.manager
    test_event1 = scan1.make_event("1.2.3.4", source=scan1.root_event)
    test_event1.make_in_scope()
    test_event2 = scan1.make_event("2.3.4.5", source=test_event1)
    test_event3 = scan1.make_event("3.4.5.6", source=test_event2)

    scan1.scope_search_distance = 1
    scan1.scope_report_distance = 0
    assert test_event1.scope_distance == 0
    manager._emit_event(test_event1)
    assert test_event1._internal == False
    assert test_event2.scope_distance == 1
    manager._emit_event(test_event2)
    assert test_event2._internal == True
    manager.events_accepted.clear()
    manager.events_distributed.clear()

    scan1.modules["json"].queue_event = lambda e: output_queue.append(e)
    scan1.modules["ipneighbor"].queue_event = lambda e: module_queue.append(e)

    # in-scope event
    assert test_event1.scope_distance == 0
    manager.distribute_event(test_event1)
    assert hash(test_event1) in manager.events_distributed
    assert test_event1 in module_queue
    assert test_event1 in output_queue
    module_queue.clear()
    output_queue.clear()
    # duplicate event
    manager.distribute_event(test_event1)
    assert test_event1 not in module_queue
    assert test_event1 in output_queue
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 1
    assert test_event2.scope_distance == 1
    manager.distribute_event(test_event2)
    assert test_event2 in module_queue
    assert test_event2 not in output_queue
    assert test_event2._internal == True
    assert test_event2._force_output == False
    assert scan1.modules["json"]._filter_event(test_event2)[0] == False
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 2
    assert test_event3.scope_distance == 2
    manager.distribute_event(test_event3)
    assert test_event3 not in module_queue
    assert test_event3 not in output_queue
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 2 and _force_output == True
    test_event3._force_output = True
    assert test_event3.scope_distance == 2
    manager.distribute_event(test_event3)
    assert test_event3 not in module_queue
    assert test_event3 in output_queue



from .crobat import crobat


class otx(crobat):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query otx.alienvault.com for subdomains"}

    base_url = "https://otx.alienvault.com"

    def request_url(self, query):
        url = f"{self.base_url}/api/v1/indicators/domain/{self.helpers.quote(query)}/passive_dns"
        return self.helpers.request(url)

    def parse_results(self, r, query):
        j = r.json()
        if isinstance(j, dict):
            for entry in j.get("passive_dns", []):
                subdomain = entry.get("hostname", "")
                if subdomain:
                    yield subdomain



import re
import random
import string

from bbot.modules.deadly.ffuf import ffuf


class ffuf_shortnames(ffuf):
    watched_events = ["URL_HINT"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["brute-force", "aggressive", "active", "web-advanced", "iis-shortnames"]
    meta = {"description": "Use ffuf in combination IIS shortnames"}

    options = {
        "wordlist": "",  # default is defined within setup function
        "wordlist_extensions": "https://raw.githubusercontent.com/danielmiessler/SecLists/master/Discovery/Web-Content/raft-small-extensions-lowercase.txt",
        "lines": 1000000,
        "max_depth": 1,
        "version": "1.5.0",
        "extensions": "",
        "ignore_redirects": False,
    }

    options_desc = {
        "wordlist": "Specify wordlist to use when finding directories",
        "wordlist_extensions": "Specify wordlist to use when making extension lists",
        "lines": "take only the first N lines from the wordlist when finding directories",
        "max_depth": "the maxium directory depth to attempt to solve",
        "version": "ffuf version",
        "extensions": "Optionally include a list of extensions to extend the keyword with (comma separated)",
        "ignore_redirects": "Explicitly ignore redirects. Enable if getting a excessive false positives.",
    }

    in_scope_only = True

    deps_ansible = [
        {
            "name": "Download ffuf",
            "unarchive": {
                "src": "https://github.com/ffuf/ffuf/releases/download/v#{BBOT_MODULES_FFUF_VERSION}/ffuf_#{BBOT_MODULES_FFUF_VERSION}_#{BBOT_OS_PLATFORM}_#{BBOT_CPU_ARCH}.tar.gz",
                "include": "ffuf",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        }
    ]

    def setup(self):
        self.sanity_canary = "".join(random.choice(string.ascii_lowercase) for i in range(10))
        wordlist = self.config.get("wordlist", "")
        if not wordlist:
            wordlist = f"{self.helpers.wordlist_dir}/ffuf_shortname_candidates.txt"
        self.wordlist = self.helpers.wordlist(wordlist)
        wordlist_extensions = self.config.get("wordlist_extensions", "")
        self.wordlist_extensions = self.helpers.wordlist(wordlist_extensions)
        self.extensions = self.config.get("extensions")
        self.ignore_redirects = self.config.get("ignore_redirects")
        return True

    def handle_event(self, event):
        filename_hint = re.sub(r"~\d", "", event.parsed.path.rsplit(".", 1)[0].split("/")[-1]).lower()

        if len(filename_hint) == 6:
            tempfile, tempfile_len = self.generate_templist(self.wordlist, prefix=filename_hint)
            self.verbose(
                f"generated temp word list of size [{str(tempfile_len)}] for filename hint: [{filename_hint}]"
            )
        else:
            tempfile = self.helpers.tempfile([filename_hint], pipe=False)
            tempfile_len = 1

        if tempfile_len > 0:
            root_stub = "/".join(event.parsed.path.split("/")[:-1])
            root_url = f"{event.parsed.scheme}://{event.parsed.netloc}{root_stub}/"

            if "shortname-file" in event.tags:
                used_extensions = []
                extension_hint = event.parsed.path.rsplit(".", 1)[1].lower().strip()

                with open(self.wordlist_extensions) as f:
                    for l in f:
                        l = l.lower().lstrip(".")
                        if l.lower().startswith(extension_hint):
                            used_extensions.append(l.strip())

                for ext in used_extensions:
                    for r in self.execute_ffuf(tempfile, event, root_url, suffix=f".{ext}"):
                        self.emit_event(r["url"], "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"])

            elif "shortname-directory" in event.tags:
                for r in self.execute_ffuf(tempfile, event, root_url):
                    self.emit_event(r["url"], "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"])



from bbot.modules.crobat import crobat


class threatminer(crobat):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query threatminer's API for subdomains",
    }

    base_url = "https://api.threatminer.org/v2"

    def request_url(self, query):
        return self.helpers.request(f"{self.base_url}/domain.php?q={self.helpers.quote(query)}&rt=5")

    def parse_results(self, r, query):
        j = r.json()
        yield from j.get("results", [])



from bbot.modules.shodan_dns import shodan_dns


class binaryedge(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the BinaryEdge API", "auth_required": True}
    options = {"api_key": "", "max_records": 1000}
    options_desc = {
        "api_key": "BinaryEdge API key",
        "max_records": "Limit results to help prevent exceeding API quota",
    }

    base_url = "https://api.binaryedge.io/v2"

    def setup(self):
        self.max_records = self.config.get("max_records", 1000)
        self.headers = {"X-Key": self.config.get("api_key", "")}
        return super().setup()

    def ping(self):
        url = f"{self.base_url}/user/subscription"
        j = self.helpers.request(url, headers=self.headers).json()
        assert j.get("requests_left", 0) > 0

    def request_url(self, query):
        # todo: host query (certs + services)
        url = f"{self.base_url}/query/domains/subdomain/{self.helpers.quote(query)}"
        return self.helpers.request(url, headers=self.headers)

    def parse_results(self, r, query):
        j = r.json()
        return j.get("events", [])



#!/usr/bin/env python3

import os
import sys
import logging
import threading
import traceback
from omegaconf import OmegaConf
from contextlib import suppress

# fix tee buffering
sys.stdout.reconfigure(line_buffering=True)

# logging
from bbot.core.logger import init_logging, get_log_level, toggle_log_level

logging_queue, logging_handlers = init_logging()

import bbot.core.errors
from bbot import __version__
from bbot.modules import module_loader
from bbot.core.configurator.args import parser
from bbot.core.helpers.logger import log_to_stderr

log = logging.getLogger("bbot.cli")
sys.stdout.reconfigure(line_buffering=True)


log_level = get_log_level()


from . import config


def main():
    err = False
    scan_name = ""

    try:
        if len(sys.argv) == 1:
            parser.print_help()
            sys.exit(1)

        # note: command line arguments are in bbot/core/configurator/args.py
        try:
            options = parser.parse_args()
        except bbot.core.errors.ArgumentError as e:
            log.warning(e)
            sys.exit(1)
            # this is intentional since sys.exit() is monkeypatched in the tests
            return

        # --version
        if options.version:
            log.stdout(__version__)
            sys.exit(0)
            return

        # --current-config
        if options.current_config:
            log.stdout(f"{OmegaConf.to_yaml(config)}")
            sys.exit(0)
            return

        log.verbose(f'Command: {" ".join(sys.argv)}')

        if options.agent_mode:
            from bbot.agent import Agent

            agent = Agent(config)
            success = agent.setup()
            if success:
                agent.start()

        else:
            from bbot.scanner import Scanner

            try:
                module_filtering = False
                if (options.list_modules or options.help_all) and not any([options.flags, options.modules]):
                    module_filtering = True
                    modules = set(module_loader.preloaded(type="scan"))
                else:
                    modules = set(options.modules)
                    # enable modules by flags
                    for m, c in module_loader.preloaded().items():
                        if m not in modules:
                            flags = c.get("flags", [])
                            if "deadly" in flags:
                                continue
                            for f in options.flags:
                                if f in flags:
                                    log.verbose(f'Enabling {m} because it has flag "{f}"')
                                    modules.add(m)

                if not options.output_modules:
                    options.output_modules = ["human", "json", "csv"]

                scanner = Scanner(
                    *options.targets,
                    modules=list(modules),
                    output_modules=options.output_modules,
                    config=config,
                    name=options.name,
                    whitelist=options.whitelist,
                    blacklist=options.blacklist,
                    strict_scope=options.strict_scope,
                    force_start=options.force,
                )

                if options.install_all_deps:
                    all_modules = list(module_loader.preloaded())
                    scanner.helpers.depsinstaller.force_deps = True
                    scanner.helpers.depsinstaller.install(*all_modules)
                    log.info("Finished installing module dependencies")
                    return

                scan_name = str(scanner.name)

                # enable modules by dependency
                # this is only a basic surface-level check
                # todo: recursive dependency graph with networkx or topological sort?
                all_modules = list(set(scanner._scan_modules + scanner._internal_modules + scanner._output_modules))
                while 1:
                    changed = False
                    dep_choices = module_loader.recommend_dependencies(all_modules)
                    if not dep_choices:
                        break
                    for event_type, deps in dep_choices.items():
                        if event_type in ("*", "all"):
                            continue
                        # skip resolving dependency if a target provides the missing type
                        if any(e.type == event_type for e in scanner.target.events):
                            continue
                        required_by = deps.get("required_by", [])
                        recommended = deps.get("recommended", [])
                        if not recommended:
                            log.hugewarning(
                                f"{len(required_by):,} modules ({','.join(required_by)}) rely on {event_type} but no modules produce it"
                            )
                        elif len(recommended) == 1:
                            log.verbose(
                                f"Enabling {next(iter(recommended))} because {len(required_by):,} modules ({','.join(required_by)}) rely on it for {event_type}"
                            )
                            all_modules = list(set(all_modules + list(recommended)))
                            scanner._scan_modules = list(set(scanner._scan_modules + list(recommended)))
                            changed = True
                        else:
                            log.hugewarning(
                                f"{len(required_by):,} modules ({','.join(required_by)}) rely on {event_type} but no enabled module produces it"
                            )
                            log.hugewarning(
                                f"Recommend enabling one or more of the following modules which produce {event_type}:"
                            )
                            for m in recommended:
                                log.warning(f" - {m}")
                    if not changed:
                        break

                # required flags
                modules = set(scanner._scan_modules)
                for m in scanner._scan_modules:
                    flags = module_loader._preloaded.get(m, {}).get("flags", [])
                    if not all(f in flags for f in options.require_flags):
                        log.verbose(
                            f"Removing {m} because it does not have the required flags: {'+'.join(options.require_flags)}"
                        )
                        modules.remove(m)

                # excluded flags
                for m in scanner._scan_modules:
                    flags = module_loader._preloaded.get(m, {}).get("flags", [])
                    if any(f in flags for f in options.exclude_flags):
                        log.verbose(f"Removing {m} because of excluded flag: {','.join(options.exclude_flags)}")
                        modules.remove(m)

                # excluded modules
                for m in options.exclude_modules:
                    if m in modules:
                        log.verbose(f"Removing {m} because it is excluded")
                        modules.remove(m)
                scanner._scan_modules = list(modules)

                log_fn = log.info
                if options.list_modules or options.help_all:
                    log_fn = log.stdout

                help_modules = list(modules)
                if module_filtering:
                    help_modules = None

                if options.help_all:
                    log_fn(parser.format_help())

                log_fn("")
                log_fn("### MODULES ###")
                log_fn("")
                for row in module_loader.modules_table(modules=help_modules).splitlines():
                    log_fn(row)

                if options.help_all:
                    log_fn("")
                    log_fn("### MODULE OPTIONS ###")
                    log_fn("")
                    for row in module_loader.modules_options_table(modules=help_modules).splitlines():
                        log_fn(row)

                if options.list_modules or options.help_all:
                    return

                module_list = module_loader.filter_modules(modules=modules)
                deadly_modules = [
                    m[0] for m in module_list if "deadly" in m[-1]["flags"] and m[0] in scanner._scan_modules
                ]
                if scanner._scan_modules and deadly_modules:
                    if not options.allow_deadly:
                        log.hugewarning(f"You enabled the following deadly modules: {','.join(deadly_modules)}")
                        log.hugewarning(f"Deadly modules are highly intrusive")
                        log.hugewarning(f"Please specify --allow-deadly to continue")
                        return

                scanner.helpers.word_cloud.load(options.load_wordcloud)

                scanner.prep()

                if not options.dry_run:
                    if not options.agent_mode and not options.yes and sys.stdin.isatty():
                        log.hugesuccess(f"Scan ready. Press enter to execute {scanner.name}")
                        input()

                    def keyboard_listen():
                        while 1:
                            keyboard_input = input()
                            if not keyboard_input:
                                toggle_log_level(logger=log)

                    keyboard_listen_thread = threading.Thread(target=keyboard_listen, daemon=True)
                    keyboard_listen_thread.start()

                    scanner.start_without_generator()

            except bbot.core.errors.ScanError as e:
                log_to_stderr(str(e), level="ERROR")
            except Exception:
                raise
            finally:
                with suppress(NameError):
                    scanner.cleanup()

    except bbot.core.errors.BBOTError as e:
        log_to_stderr(f"{e} (--debug for details)", level="ERROR")
        if log_level <= logging.DEBUG:
            log_to_stderr(traceback.format_exc(), level="DEBUG")
        err = True

    except Exception:
        log_to_stderr(f"Encountered unknown error: {traceback.format_exc()}", level="ERROR")
        err = True

    except KeyboardInterrupt:
        msg = "Interrupted"
        if scan_name:
            msg = f"You killed {scan_name}"
        log_to_stderr(msg, level="ERROR")
        err = True

    finally:
        # save word cloud
        with suppress(Exception):
            save_success, filename = scanner.helpers.word_cloud.save(options.save_wordcloud)
            if save_success:
                log_to_stderr(f"Saved word cloud ({len(scanner.helpers.word_cloud):,} words) to {filename}")
        # remove output directory if empty
        with suppress(Exception):
            scanner.home.rmdir()
        if err:
            os._exit(1)

        # debug troublesome modules
        """
        from time import sleep
        while 1:
            scanner.manager.modules_status(_log=True)
            sleep(1)
        """


if __name__ == "__main__":
    main()



from bbot.modules.base import BaseModule


class crobat(BaseModule):
    """
    A typical free API-based subdomain enumeration module
    Inherited by several other modules including sublist3r, dnsdumpster, etc.
    """

    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query Project Crobat for subdomains"}

    base_url = "https://sonar.omnisint.io"

    # this helps combat rate limiting by ensuring that a query doesn't execute
    # until the queue is ready to receive its results
    _qsize = 1

    def setup(self):
        self.processed = set()
        return True

    def filter_event(self, event):
        """
        Accept DNS_NAMEs that are either directly targets, or indirectly
        in scope by resolving to in-scope IPs.

        Kill wildcards with fire.

        This filter_event is used across many modules
        """
        query = self.make_query(event)
        if self.already_processed(query):
            return False, "Event was already processed"
        if not "target" in event.tags:
            if "unresolved" in event.tags:
                return False, "Event is unresolved"
            if any(t.startswith("cloud-") for t in event.tags):
                return False, "Event is a cloud resource and not a direct target"
        for domain, wildcard_rdtypes in self.helpers.is_wildcard_domain(query).items():
            if any(t in wildcard_rdtypes for t in ("A", "AAAA", "CNAME")):
                return False, "Event is a wildcard domain"
        if any(t in event.tags for t in ("a-error", "aaaa-error")):
            return False, "Event has a DNS resolution error"
        self.processed.add(hash(query))
        return True

    def already_processed(self, hostname):
        for parent in self.helpers.domain_parents(hostname, include_self=True):
            if hash(parent) in self.processed:
                return True
        return False

    def abort_if(self, event):
        # this helps weed out unwanted results when scanning IP_RANGES and wildcard domains
        return "in-scope" not in event.tags or "wildcard" in event.tags

    def handle_event(self, event):
        query = self.make_query(event)
        results = self.query(query)
        if results:
            for hostname in results:
                if not hostname == event:
                    self.emit_event(hostname, "DNS_NAME", event, abort_if=self.abort_if)

    def request_url(self, query):
        url = f"{self.base_url}/subdomains/{self.helpers.quote(query)}"
        return self.helpers.request(url)

    def make_query(self, event):
        if "target" in event.tags:
            query = str(event.data)
        else:
            query = self.helpers.parent_domain(event.data).lower()
        return ".".join([s for s in query.split(".") if s != "_wildcard"])

    def parse_results(self, r, query=None):
        json = r.json()
        if json:
            for hostname in json:
                yield hostname

    def query(self, query, parse_fn=None, request_fn=None):
        if parse_fn is None:
            parse_fn = self.parse_results
        if request_fn is None:
            request_fn = self.request_url
        try:
            results = list(parse_fn(request_fn(query), query))
            if results:
                return results
            self.debug(f'No results for "{query}"')
        except Exception:
            self.verbose(f"Error retrieving results for {query}")
            self.trace()



import re
from collections import OrderedDict

# for extracting words from strings
word_regexes = [
    re.compile(r, re.I)
    for r in [
        # alpha
        r"[a-z]{3,}",
        # alphanum
        r"[a-z0-9]{3,}",
        # alpha, dash
        r"[a-z][a-z-]+[a-z]",
        # alpha, underscore
        r"[a-z][a-z_]+[a-z]",
    ]
]


word_regex = re.compile(r"[^\d\W_]+")
word_num_regex = re.compile(r"[^\W_]+")
num_regex = re.compile(r"\d+")
_ipv6_regex = r"[A-F0-9:]*:[A-F0-9:]*:[A-F0-9:]*"
ipv6_regex = re.compile(_ipv6_regex, re.I)
_dns_name_regex = r"(?:(?:[\w-]+)\.)+(?:[^\W_0-9]{2,20})"
_hostname_regex = re.compile(r"^[\w-]+$")
_email_regex = r"(?:[^\W_][\w\-\.\+]{,100})@(?:\w[\w\-\._]{,100})\.(?:[^\W_0-9]{2,8})"
email_regex = re.compile(_email_regex, re.I)

event_type_regexes = OrderedDict(
    [
        (k, tuple(re.compile(r, re.I) for r in regexes))
        for k, regexes in [
            (
                "DNS_NAME",
                (r"^" + _dns_name_regex + r"$",),
            ),
            (
                "EMAIL_ADDRESS",
                (r"^" + _email_regex + r"$",),
            ),
            (
                "OPEN_TCP_PORT",
                (
                    r"^((?:\w|\w[\w\-]*\w)[\.]?)+(?:\w[\w\-]*\w|\w):[0-9]{1,5}$",
                    r"^\[" + _ipv6_regex + r"\]:[0-9]{1,5}$",
                ),
            ),
            (
                "URL",
                (
                    r"https?://((?:\w|\w[\w\-]*\w)[\.]?)+(?:\w[\w\-]*\w|\w)(?::[0-9]{1,5})?.*$",
                    r"https?://\[" + _ipv6_regex + r"\](?::[0-9]{1,5})?.*$",
                ),
            ),
        ]
    ]
)

event_id_regex = re.compile(r"[0-9a-f]{40}:[A-Z0-9_]+")
dns_name_regex = re.compile(_dns_name_regex, re.I)
scan_name_regex = re.compile(r"[a-z]{3,20}_[a-z]{3,20}")



import re
import json
import requests

from bbot.modules.base import BaseModule
from bbot.core.helpers.misc import tldextract


class subdomain_hijack(BaseModule):
    flags = ["subdomain-hijack", "subdomain-enum", "cloud-enum", "safe", "active"]
    watched_events = ["DNS_NAME"]
    produced_events = ["FINDING"]
    meta = {"description": "Detect hijackable subdomains"}
    options = {
        "fingerprints": "https://raw.githubusercontent.com/blacklanternsecurity/can-i-take-over-xyz/master/fingerprints.json"
    }
    options_desc = {"fingerprints": "URL or path to fingerprints.json"}
    scope_distance_modifier = 2
    max_event_handlers = 5

    def setup(self):
        fingerprints_url = self.config.get("fingerprints")
        fingerprints_file = self.helpers.wordlist(fingerprints_url)
        with open(fingerprints_file) as f:
            fingerprints = json.load(f)
        self.fingerprints = []
        for f in fingerprints:
            try:
                f = Fingerprint(f)
            except Exception as e:
                self.warning(f"Error instantiating fingerprint: {e}")
                continue
            if not (f.domains and f.vulnerable and f.fingerprint and f.cicd_pass):
                self.debug(f"Skipping fingerprint: {f}")
                continue
            self.debug(f"Processed fingerprint: {f}")
            self.fingerprints.append(f)
        if not self.fingerprints:
            return None, "No valid fingerprints"
        self.debug(f"Successfully processed {len(self.fingerprints):,} fingerprints")
        return True

    def filter_event(self, event):
        return "subdomain" in event.tags

    def handle_event(self, event):
        hijackable, reason = self.check_subdomain(event.data)
        if hijackable:
            source_hosts = []
            e = event
            while 1:
                host = getattr(e, "host", "")
                if host:
                    if host not in source_hosts:
                        source_hosts.append(host)
                    e = e.source
                else:
                    break

            url = f"https://{event.host}"
            description = f"Hijackable Subdomain"
            if source_hosts:
                source_hosts_str = " -> ".join([str(s) for s in source_hosts[::-1]])
                description += f" ({source_hosts_str})"
            self.emit_event({"host": event.host, "url": url, "description": description}, "FINDING", source=event)
        else:
            self.debug(reason)

    def check_subdomain(self, subdomain):
        for f in self.fingerprints:
            for domain in f.domains:
                if self.helpers.host_in_host(subdomain, domain):
                    url = f"https://{subdomain}"
                    try:
                        r = self.helpers.request(url, raise_error=True)
                        text = getattr(r, "text", "")
                        if not f.nxdomain and f.fingerprint_regex.findall(text):
                            return True, subdomain
                    except requests.exceptions.RequestException as e:
                        if f.nxdomain and "Name or service not known" in str(e):
                            return True, f"{subdomain} --> NXDOMAIN"
        return False, f'Subdomain "{subdomain}" not hijackable'


class Fingerprint:
    def __init__(self, fingerprint):
        assert isinstance(fingerprint, dict), "fingerprint must be a dictionary"
        self.engine = fingerprint.get("service")
        self.cnames = fingerprint.get("cname", [])
        self.domains = list(set([tldextract(c).registered_domain for c in self.cnames]))
        self.nxdomain = fingerprint.get("nxdomain", False)
        self.vulnerable = fingerprint.get("vulnerable", False)
        self.fingerprint = fingerprint.get("fingerprint", "")
        self.cicd_pass = fingerprint.get("cicd_pass", False)
        try:
            self.fingerprint_regex = re.compile(self.fingerprint, re.MULTILINE)
        except re.error:
            self.fingerprint_regex = re.compile(re.escape(self.fingerprint), re.MULTILINE)

    def __str__(self):
        return f"{self.engine}: {self.fingerprint} (cnames: {self.cnames}, vulnerable: {self.vulnerable}, cicd_pass: {self.cicd_pass})"



from bbot.modules.base import BaseModule


class bucket_aws(BaseModule):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum"]
    meta = {"description": "Check for S3 buckets related to target"}
    options = {"max_threads": 10, "permutations": False}
    options_desc = {
        "max_threads": "Maximum number of threads for HTTP requests",
        "permutations": "Whether to try permutations",
    }
    scope_distance_modifier = 1

    cloud_helper_name = "aws"
    delimiters = ("", ".", "-")
    base_domains = ["s3.amazonaws.com"]
    regions = [None]
    supports_open_check = True

    def setup(self):
        self.buckets_tried = set()
        self.cloud_helper = getattr(self.helpers.cloud, self.cloud_helper_name)
        self.permutations = self.config.get("permutations", False)
        return True

    def filter_event(self, event):
        if event.type == "STORAGE_BUCKET":
            return True
        dns_name_scope_distance = max(0, self.max_scope_distance - 2)
        if event.scope_distance <= dns_name_scope_distance:
            return True
        return False

    def handle_event(self, event):
        if event.type == "DNS_NAME":
            self.handle_dns_name(event)
        elif event.type == "STORAGE_BUCKET" and self.supports_open_check:
            self.handle_storage_bucket(event)

    def handle_dns_name(self, event):
        buckets = set()
        base = event.data
        stem = self.helpers.domain_stem(base)
        for b in [base, stem]:
            split = b.split(".")
            for d in self.delimiters:
                bucket_name = d.join(split)
                if self.valid_bucket_name(bucket_name):
                    buckets.add(bucket_name)

        if self.permutations:
            for b in [base, stem]:
                for mutation in self.helpers.word_cloud.mutations(b):
                    for d in self.delimiters:
                        bucket_name = d.join(mutation)
                        if self.valid_bucket_name(bucket_name):
                            buckets.add(bucket_name)

        futures = {}
        for base_domain in self.base_domains:
            for region in self.regions:
                for bucket_name in buckets:
                    url = self.build_url(bucket_name, base_domain, region)
                    future = self.submit_task(self.check_bucket_exists, bucket_name, url)
                    futures[future] = (bucket_name, url)
        for future in self.helpers.as_completed(futures):
            bucket_name, url = futures[future]
            existent_bucket, tags = future.result()
            if existent_bucket:
                self.emit_event({"name": bucket_name, "url": url}, "STORAGE_BUCKET", source=event, tags=tags)

    def handle_storage_bucket(self, event):
        url = event.data["url"]
        bucket_name = event.data["name"]
        description, tags = self.check_bucket_open(bucket_name, url)
        if description:
            self.emit_event(
                {"host": event.host, "url": url, "description": description}, "FINDING", source=event, tags=tags
            )

    def check_bucket_exists(self, bucket_name, url):
        response = self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        existent_bucket = status_code != 404
        return (existent_bucket, tags)

    def check_bucket_open(self, bucket_name, url):
        response = self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        content = getattr(response, "text", "")
        open_bucket = status_code == 200 and "Contents" in content
        msg = ""
        if open_bucket:
            msg = "Open storage bucket"
        return (msg, tags)

    def valid_bucket_name(self, bucket_name):
        valid = self.cloud_helper.is_valid_bucket(bucket_name)
        if valid and not self.helpers.is_ip(bucket_name):
            bucket_hash = hash(bucket_name)
            if not bucket_hash in self.buckets_tried:
                self.buckets_tried.add(bucket_hash)
                return True
        return False

    def build_url(self, bucket_name, base_domain, region):
        return f"https://{bucket_name}.{base_domain}"

    def gen_tags_exists(self, response):
        return set()

    def gen_tags_open(self, response):
        return set()



import requests_mock
from abc import abstractmethod
from omegaconf import OmegaConf


class MockHelper:
    targets = ["blacklanternsecurity.com"]
    config_overrides = {}
    additional_modules = []

    def __init__(self, config, bbot_scanner, *args):
        self.name = self.__class__.__name__.lower()
        self.config = OmegaConf.merge(config, OmegaConf.create(self.config_overrides))
        self.scan = bbot_scanner(
            *self.targets, modules=[self.name] + self.additional_modules, name=f"{self.name}_test", config=self.config
        )
        self.patch_scan(self.scan)
        self.scan.prep()
        self.module = self.scan.modules[self.name]
        self.setup()

    def patch_scan(self, scan):
        return

    def setup(self):
        pass

    def run(self):
        events = list(self.scan.start())
        for e in events:
            print(e)
        events = [e for e in events if e.module == self.module]
        assert self.check_events(events)

    @abstractmethod
    def check_events(self, events):
        raise NotImplementedError


class RequestMockHelper(MockHelper):
    @abstractmethod
    def mock_args(self):
        raise NotImplementedError

    def register_uri(self, uri, method="GET", **kwargs):
        self.m.register_uri(method, uri, **kwargs)

    def run(self):
        with requests_mock.Mocker() as m:
            self.m = m
            self.mock_args()
            events = list(e for e in self.scan.start() if e.module == self.module)
            for x in events:
                print(x)
            assert self.check_events(events)


class HttpxMockHelper(MockHelper):
    targets = ["http://127.0.0.1:8888/"]

    def __init__(self, config, bbot_scanner, bbot_httpserver):
        self.bbot_httpserver = bbot_httpserver
        super().__init__(config, bbot_scanner)
        self.mock_args()

    @abstractmethod
    def mock_args(self):
        raise NotImplementedError

    def set_expect_requests(self, expect_args={}, respond_args={}):
        if "uri" not in expect_args:
            expect_args["uri"] = "/"
        self.bbot_httpserver.expect_request(**expect_args).respond_with_data(**respond_args)


def tempwordlist(content):
    tmp_path = "/tmp/.bbot_test/"
    from bbot.core.helpers.misc import rand_string, mkdir

    mkdir(tmp_path)
    filename = f"{tmp_path}{rand_string(8)}"
    with open(filename, "w", errors="ignore") as f:
        for c in content:
            line = f"{c}\n"
            f.write(line)
    return filename



from bbot.modules.base import BaseModule
from bbot.core.errors import InteractshError


class host_header(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "web-advanced"]
    meta = {"description": "Try common HTTP Host header spoofing techniques"}

    in_scope_only = True

    deps_apt = ["curl"]

    def setup(self):
        self.interactsh_subdomain_tags = {}
        if self.scan.config.get("interactsh_disable", False) == False:
            try:
                self.interactsh_instance = self.helpers.interactsh()
                self.interactsh_domain = self.interactsh_instance.register(callback=self.interactsh_callback)
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")
                return False
        return True

    def interactsh_callback(self, r):
        full_id = r.get("full-id", None)
        if full_id:
            if "." in full_id:
                match = self.interactsh_subdomain_tags.get(full_id.split(".")[0])
                if match is None:
                    return
                matched_event = match[0]
                matched_technique = match[1]

                self.emit_event(
                    {
                        "host": str(matched_event.host),
                        "url": matched_event.data["url"],
                        "description": f"Spoofed Host header ({matched_technique}) [{r.get('protocol').upper()}] interaction",
                    },
                    "FINDING",
                    matched_event,
                )
            else:
                # this is likely caused by something trying to resolve the base domain first and can be ignored
                self.debug("skipping results because subdomain tag was missing")

    def finish(self):
        try:
            for r in self.interactsh_instance.poll():
                self.interactsh_callback(r)
        except InteractshError as e:
            self.debug(f"Error in interact.sh: {e}")

    def cleanup(self):
        try:
            self.interactsh_instance.deregister()
            self.debug(
                f"successfully deregistered interactsh session with correlation_id {self.interactsh_instance.correlation_id}"
            )
        except InteractshError as e:
            self.warning(f"Interactsh failure: {e}")

    def handle_event(self, event):
        # get any set-cookie responses from the response and add them to the request

        added_cookies = {}

        for k, v in event.data["header-dict"].items():
            if k.lower() == "set-cookie":
                cookie_string = v
                cookie_split = cookie_string.split("=")
                added_cookies = {cookie_split[0]: cookie_split[1]}

        domain_reflections = []

        # host header replacement
        technique_description = "standard"
        self.debug(f"Performing {technique_description} case")
        subdomain_tag = self.helpers.rand_string(4, digits=False)
        self.interactsh_subdomain_tags[subdomain_tag] = (event, technique_description)
        output = self.helpers.curl(
            url=event.data["url"],
            headers={"Host": f"{subdomain_tag}.{self.interactsh_domain}"},
            ignore_bbot_global_settings=True,
            cookies=added_cookies,
        )

        if self.interactsh_domain in output:
            domain_reflections.append(technique_description)

        # absolute URL / Host header transposition
        technique_description = "absolute URL transposition"
        self.debug(f"Performing {technique_description} case")
        subdomain_tag = self.helpers.rand_string(4, digits=False)
        self.interactsh_subdomain_tags[subdomain_tag] = (event, technique_description)
        output = self.helpers.curl(
            url=event.data["url"],
            headers={"Host": f"{subdomain_tag}.{self.interactsh_domain}"},
            path_override=event.data["url"],
            cookies=added_cookies,
        )

        if self.interactsh_domain in output:
            domain_reflections.append(technique_description)

        # duplicate host header tolerance
        technique_description = "duplicate host header tolerance"
        output = self.helpers.curl(
            url=event.data["url"],
            # Sending a blank HOST first as a hack to trick curl. This makes it no longer an "internal header", thereby allowing for duplicates
            # The fact that it's accepting two host headers is rare enough to note on its own, and not too noisy. Having the 3rd header be an interactsh would result in false negatives for the slightly less interesting cases.
            headers={"Host": ["", str(event.host), str(event.host)]},
            cookies=added_cookies,
            head_mode=True,
        )

        split_output = output.split("\n")
        if " 4" in split_output:
            self.emit_event(
                {
                    "host": str(event.host),
                    "url": event.data["url"],
                    "description": f"Duplicate Host Header Tolerated",
                },
                "FINDING",
                event,
            )

        # Host Header Overrides

        technique_description = "host override headers"
        self.debug(f"Performing {technique_description} case")
        subdomain_tag = self.helpers.rand_string(4, digits=False)
        self.interactsh_subdomain_tags[subdomain_tag] = (event, technique_description)

        override_headers_list = [
            "X-Host",
            "X-Forwarded-Server",
            "X-Forwarded-Host",
            "X-Original-Host",
            "X-Forwarded-For",
            "X-Host",
            "X-HTTP-Host-Override",
            "Forwarded",
        ]
        override_headers = {}
        for oh in override_headers_list:
            override_headers[oh] = f"{subdomain_tag}.{self.interactsh_domain}"

        output = self.helpers.curl(
            url=event.data["url"],
            headers=override_headers,
            cookies=added_cookies,
        )
        if self.interactsh_domain in output:
            domain_reflections.append(technique_description)

        # emit all the domain reflections we found
        for dr in domain_reflections:
            self.emit_event(
                {
                    "host": str(event.host),
                    "url": event.data["url"],
                    "description": f"Possible Host header injection. Injection technique: {dr}",
                },
                "FINDING",
                event,
            )



from .crobat import crobat


class sublist3r(crobat):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query sublist3r's API for subdomains",
    }

    base_url = "https://api.sublist3r.com/search.php"

    def request_url(self, query):
        return self.helpers.request(f"{self.base_url}?domain={query}")

    def parse_results(self, r, query):
        json = r.json()
        if json:
            for hostname in json:
                yield hostname



from bbot.modules.base import BaseModule
from bbot.core.errors import HttpCompareError, ScanCancelledError


class header_brute(BaseModule):
    """
    Inspired by https://github.com/PortSwigger/param-miner
    """

    watched_events = ["URL"]
    produced_events = ["FINDING"]
    flags = ["brute-force", "active", "aggressive", "slow", "web-paramminer"]
    meta = {"description": "Check for common HTTP header parameters"}
    options = {"wordlist": "https://raw.githubusercontent.com/PortSwigger/param-miner/master/resources/headers"}
    options_desc = {"wordlist": "Define the wordlist to be used to derive headers"}
    scanned_hosts = []
    header_blacklist = [
        "content-length",
        "expect",
        "accept-encoding",
        "transfer-encoding",
        "connection",
        "if-match",
        "if-modified-since",
        "if-none-match",
        "if-unmodified-since",
    ]
    max_event_handlers = 12
    in_scope_only = True
    compare_mode = "header"

    def setup(self):
        wordlist_url = self.config.get("wordlist", "")
        self.wordlist = self.helpers.wordlist(wordlist_url)
        return True

    def rand_string(self, *args, **kwargs):
        return self.helpers.rand_string(*args, **kwargs)

    def handle_event(self, event):
        url = event.data
        try:
            compare_helper = self.helpers.http_compare(url)
        except HttpCompareError as e:
            self.debug(e)
            return
        batch_size = self.count_test(url)
        if batch_size == None or batch_size <= 0:
            self.debug(f"Failed to get baseline max {self.compare_mode} count, aborting")
            return
        self.debug(f"Resolved batch_size at {str(batch_size)}")

        if compare_helper.canary_check(url, mode=self.compare_mode) == False:
            self.warning(f'Aborting "{url}" due to failed canary check')
            return

        fl = [h.strip().lower() for h in self.helpers.read_file(self.wordlist)]

        wordlist_cleaned = list(filter(self.clean_list, fl))

        results = set()
        abort_threshold = 25
        try:
            for group in self.helpers.grouper(wordlist_cleaned, batch_size):
                for result, reasons, reflection in self.binary_search(compare_helper, url, group):
                    results.add((result, ",".join(reasons), reflection))
                    if len(results) >= abort_threshold:
                        self.warning(
                            f"Abort threshold ({abort_threshold}) reached, too many {self.compare_mode}s found"
                        )
                        results.clear()
                        assert False
        except ScanCancelledError:
            return
        except AssertionError:
            pass

        for result, reasons, reflection in results:
            tags = []
            if reflection:
                tags = ["http_reflection"]
            description = f"[{self.compare_mode.upper()}_BRUTE] {self.compare_mode.capitalize()}: [{result}] Reasons: [{reasons}]"
            self.emit_event(
                {"host": str(event.host), "url": url, "description": description},
                "FINDING",
                event,
                tags=tags,
            )

    def count_test(self, url):
        baseline = self.helpers.request(url)
        if baseline is None:
            return
        if str(baseline.status_code)[0] in ("4", "5"):
            return
        for count, args, kwargs in self.gen_count_args(url):
            r = self.helpers.request(*args, **kwargs)
            if r is not None and not ((str(r.status_code)[0] in ("4", "5"))):
                return count

    def gen_count_args(self, url):
        header_count = 95
        while 1:
            if header_count < 0:
                break
            fake_headers = {}
            for i in range(0, header_count):
                fake_headers[self.rand_string(14)] = self.rand_string(14)
            yield header_count, (url,), {"headers": fake_headers}
            header_count -= 5

    def clean_list(self, header):
        if (len(header) > 0) and ("%" not in header) and (header not in self.header_blacklist):
            return True
        return False

    def binary_search(self, compare_helper, url, group, reasons=None, reflection=False):
        if reasons is None:
            reasons = []
        self.debug(f"Entering recursive binary_search with {len(group):,} sized group")
        if len(group) == 1:
            yield group[0], reasons, reflection
        elif len(group) > 1:
            for group_slice in self.helpers.split_list(group):
                match, reasons, reflection, subject_response = self.check_batch(compare_helper, url, group_slice)
                if match == False:
                    yield from self.binary_search(compare_helper, url, group_slice, reasons, reflection)
        else:
            self.warning(f"Submitted group of size 0 to binary_search()")

    def check_batch(self, compare_helper, url, header_list):
        if self.scan.stopping:
            raise ScanCancelledError()
        rand = self.rand_string()
        test_headers = {}
        for header in header_list:
            test_headers[header] = rand
        return compare_helper.compare(url, headers=test_headers, check_reflection=(len(header_list) == 1))



from .crobat import crobat


class anubisdb(crobat):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query jldc.me's database for subdomains"}

    base_url = "https://jldc.me/anubis/subdomains"

    def request_url(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}"
        return self.helpers.request(url)

    def abort_if(self, event):
        # abort if dns name is unresolved
        return not "resolved" in event.tags or super().abort_if(event)

    def parse_results(self, r, query):
        results = set()
        json = r.json()
        if json:
            for hostname in json:
                hostname = str(hostname).lower()
                if hostname.endswith(f".{query}"):
                    results.add(hostname)
        return results



from .base import BaseCloudProvider


class Azure(BaseCloudProvider):
    domains = [
        "windows.net",
        "azure.com",
        "azmk8s.io",
        "azure-api.net",
        "azure-mobile.net",
        "azurecontainer.io",
        "azurecr.io",
        "azureedge.net",
        "azurefd.net",
        "azurewebsites.net",
        "cloudapp.net",
        "onmicrosoft.com",
        "trafficmanager.net",
        "vault.azure.net",
        "visualstudio.com",
        "vo.msecnd.net",
    ]

    bucket_name_regex = r"[a-z0-9][a-z0-9-_\.]{1,61}[a-z0-9]"
    regexes = {"STORAGE_BUCKET": [r"(%[a-f0-9]{2})?(" + bucket_name_regex + r")\.(blob\.core\.windows\.net)"]}



from ..bbot_fixtures import *


def test_scan(
    patch_ansible,
    patch_commands,
    events,
    bbot_config,
    helpers,
    neograph,
    websocketapp,
    monkeypatch,
    bbot_scanner,
):
    scan0 = bbot_scanner(
        "8.8.8.8/31", "evilcorp.com", blacklist=["8.8.8.8/28", "www.evilcorp.com"], config=bbot_config
    )
    assert scan0.whitelisted("8.8.8.8")
    assert scan0.whitelisted("8.8.8.9")
    assert scan0.blacklisted("8.8.8.15")
    assert not scan0.blacklisted("8.8.8.16")
    assert scan0.blacklisted("8.8.8.8/30")
    assert not scan0.blacklisted("8.8.8.8/27")
    assert not scan0.in_scope("8.8.8.8")
    assert scan0.whitelisted("api.evilcorp.com")
    assert scan0.whitelisted("www.evilcorp.com")
    assert not scan0.blacklisted("api.evilcorp.com")
    assert scan0.blacklisted("asdf.www.evilcorp.com")
    assert scan0.in_scope("test.api.evilcorp.com")
    assert not scan0.in_scope("test.www.evilcorp.com")
    assert not scan0.in_scope("www.evilcorp.co.uk")

    scan1 = bbot_scanner("8.8.8.8", whitelist=["8.8.4.4"], config=bbot_config)
    assert not scan1.blacklisted("8.8.8.8")
    assert not scan1.blacklisted("8.8.4.4")
    assert not scan1.whitelisted("8.8.8.8")
    assert scan1.whitelisted("8.8.4.4")
    assert scan1.in_scope("8.8.4.4")
    assert not scan1.in_scope("8.8.8.8")

    scan2 = bbot_scanner("8.8.8.8", config=bbot_config)
    assert not scan2.blacklisted("8.8.8.8")
    assert not scan2.blacklisted("8.8.4.4")
    assert scan2.whitelisted("8.8.8.8")
    assert not scan2.whitelisted("8.8.4.4")
    assert scan2.in_scope("8.8.8.8")
    assert not scan2.in_scope("8.8.4.4")

    scan3 = bbot_scanner(
        "127.0.0.0/30",
        "127.0.0.2:8443",
        "https://localhost",
        "[::1]:80",
        "http://[::1]:8080",
        modules=["ipneighbor"],
        output_modules=list(available_output_modules),
        config=bbot_config,
        blacklist=["http://127.0.0.3:8000/asdf"],
        whitelist=["127.0.0.0/29"],
    )
    patch_commands(scan3)
    patch_ansible(scan3)
    assert "targets" in scan3.json
    assert "127.0.0.3" in scan3.target
    assert "127.0.0.4" not in scan3.target
    assert "127.0.0.4" in scan3.whitelist
    assert scan3.whitelisted("127.0.0.4")
    assert "127.0.0.3" in scan3.blacklist
    assert scan3.blacklisted("127.0.0.3")
    assert scan3.in_scope("127.0.0.1")
    assert not scan3.in_scope("127.0.0.3")
    scan3.prep()
    monkeypatch.setattr(scan3.modules["websocket"], "ws", websocketapp())
    scan3.start()
